///////////////////////////////////////////////////////////////////////
// Matroska plugin for VirtualDub 1.8.8+
// Copyright (C) 2010-2012 fccHandler
// Copyright (C) 1998-2012 Avery Lee
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////

// IVDXAudioSource implementation

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <vfw.h>
#include <ks.h>
#include <ksmedia.h>

#include "VDInputDriver.h"
#include "VDUnknown.h"
#include "MkvFile.h"

#include "DataVector.h"

#include "InputFileMKV.h"
#include "MatroskaTrack.h"
#include "AudioSourceMKV.h"


///////////////////////////////////////////////////////////////////////
//
//                      AudioSourceMKV
//
///////////////////////////////////////////////////////////////////////


AudioSourceMKV::AudioSourceMKV(InputFileMKV *const pp, MatroskaTrack *const mt)
	: parentPtr(pp)
	, audio(mt)
{
	parentPtr->AddRef();

	errmode = kErrorModeDecodeAnyway;

	if (InputFileMKV::IsPCM(*audio))
	{
		pStreamBuf = new unsigned char[audio->dwMaxSample];
	}
	else
	{
		pStreamBuf = NULL;
	}

	dwStreamBlock  = 0;
	iStreamPos     = 0;
}


AudioSourceMKV::~AudioSourceMKV()
{
	delete[] pStreamBuf;

	parentPtr->Release();
}


int AudioSourceMKV::AddRef() {
	return vdxunknown<IVDXStreamSource>::AddRef();
}


int AudioSourceMKV::Release() {
	return vdxunknown<IVDXStreamSource>::Release();
}


void *VDXAPIENTRY AudioSourceMKV::AsInterface(uint32 iid) {
	if (iid == IVDXAudioSource::kIID)
		return static_cast<IVDXAudioSource *>(this);

	if (iid == IVDXStreamSourceV3::kIID)
		return static_cast<IVDXStreamSourceV3 *>(this);

	return vdxunknown<IVDXStreamSource>::AsInterface(iid);
}


void VDXAPIENTRY AudioSourceMKV::GetStreamSourceInfo(VDXStreamSourceInfo& srcInfo)
{
	switch (audio->iCodecID) {
	
	case A_MS_ACM:
	{
		// Prefer the header fields for ACM audio
		const WAVEFORMAT *wf = (WAVEFORMAT *)audio->pFormat;

		srcInfo.mSampleRate.mNumerator   = wf->nAvgBytesPerSec;
		srcInfo.mSampleRate.mDenominator = wf->nBlockAlign;
		if (InputFileMKV::IsPCM(*audio))
		{
			srcInfo.mSampleCount = audio->ui64TotalBytes / wf->nBlockAlign;
		}
		else
		{
			srcInfo.mSampleCount = audio->dwActualLength;
		}
		break;
	}

	case A_PCM_INT_LIT:
	case A_PCM_INT_BIG:
	{
		const WAVEFORMAT *wf = (WAVEFORMAT *)audio->pFormat;

		srcInfo.mSampleRate.mNumerator   = wf->nAvgBytesPerSec;
		srcInfo.mSampleRate.mDenominator = wf->nBlockAlign;
		srcInfo.mSampleCount = audio->ui64TotalBytes / wf->nBlockAlign;
		break;
	}
	
	default:
		srcInfo.mSampleRate.mNumerator   = audio->dwRate;
		srcInfo.mSampleRate.mDenominator = audio->dwScale;
		srcInfo.mSampleCount = audio->dwActualLength;
		break;
	}

	srcInfo.mPixelAspectRatio.mNumerator = 1;
	srcInfo.mPixelAspectRatio.mDenominator = 1;
}


void VDXAPIENTRY AudioSourceMKV::GetStreamSourceInfoV3(VDXStreamSourceInfoV3& info)
{
	GetStreamSourceInfo(info.mInfo);

	if (audio->dwFlags & MKV_HAS_VSB)
		info.mFlags  = VDXStreamSourceInfoV3::kFlagVariableSizeSamples;
	else
		info.mFlags  = 0;

	info.mfccHandler = 0;
}


void AudioSourceMKV::PCMByteSwap(void *lpBuffer, int len)
{
	if (lpBuffer != NULL)
	{
		unsigned char *src = (unsigned char *)lpBuffer;
		while (len > 1)
		{
			unsigned char byte = src[0];
			src[0] = src[1];
			src[1] = byte;
			src += 2;
			len -= 2;
		}
	}
}


int AudioSourceMKV::ReadStream(void *lpBuffer, __int64 pos, int len)
{
	unsigned char *dst = (unsigned char *)lpBuffer;

	if (pStreamBuf != NULL && lpBuffer != NULL && len > 0)
	{
		if (pos < iStreamPos)
		{
			dwStreamBlock = 0;
			iStreamPos = 0;
		}

		// Locate the wanted block
		while (dwStreamBlock < audio->dwActualLength)
		{
			int iSize = audio->iCompSettingsLen + (audio->sample_list[dwStreamBlock].size & (~MKV_KEYFRAME));
			if (pos < (iStreamPos + iSize)) break;

			iStreamPos += iSize;
			++dwStreamBlock;
		}

		pos -= iStreamPos;

		while (dwStreamBlock < audio->dwActualLength)
		{
			int iSize = audio->sample_list[dwStreamBlock].size & (~MKV_KEYFRAME);

			if (audio->iCompSettingsLen > 0)
			{
				memcpy(pStreamBuf, audio->pCompSettings, audio->iCompSettingsLen);
			}

			if (parentPtr->ReadData(pStreamBuf + audio->iCompSettingsLen, audio->sample_list[dwStreamBlock].pos, iSize) != iSize) break;
			iSize += audio->iCompSettingsLen;
			
			int tc = iSize - (int)pos;
			if (tc > len) tc = len;
			
			memcpy(dst, pStreamBuf + (int)pos, tc);

			len -= tc;
			dst += tc;
			if (len <= 0) break;

			++dwStreamBlock;
			iStreamPos += iSize;
			pos = 0;
		}
	}
	
	return (int)(dst - (unsigned char *)lpBuffer);
}


bool AudioSourceMKV::ReadPCM(sint64 lStart64, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lBytesRead, uint32 *lSamplesRead)
{
	int bytes;

	const DWORD ba = ((WAVEFORMAT *)audio->pFormat)->nBlockAlign;

	// buffer size query or buffer too small?
	if (lpBuffer == NULL || cbBuffer < ba)
	{
		if (lCount < 1) lCount = 1;
		if (lBytesRead) *lBytesRead = lCount * ba;
		if (lSamplesRead) *lSamplesRead = lCount;
		return (lpBuffer == NULL);
	}

	bytes = ReadStream(lpBuffer, lStart64 * ba, lCount * ba);

	if (audio->iCodecID == A_PCM_INT_BIG)
	{
		PCMByteSwap(lpBuffer, bytes);
	}

	if (lBytesRead) *lBytesRead = bytes;
	if (lSamplesRead) *lSamplesRead = bytes / ba;

	return true;
}


bool AudioSourceMKV::Read(sint64 lStart64, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lBytesRead, uint32 *lSamplesRead)
{
	const MKVSampleInfo *msi;
	unsigned char *dst;
	DWORD dwStart;
	DWORD actual;
	DWORD i;

	unsigned int profile;
	unsigned int chcode;
	unsigned int srcode;
	unsigned int iADTSlen;

	if (InputFileMKV::IsPCM(*audio))
	{
		return ReadPCM(lStart64, lCount, lpBuffer, cbBuffer, lBytesRead, lSamplesRead);
	}

	dwStart = (DWORD)lStart64;

	// Start out of range?
	if (dwStart >= audio->dwActualLength)
	{
		if (lBytesRead) *lBytesRead = 0;
		if (lSamplesRead) *lSamplesRead = 0;
		return true;
	}

	// Sample count out of range?
	if ((dwStart + lCount) > audio->dwActualLength)
	{
		lCount = audio->dwActualLength - dwStart;
	}

	// Wrap AAC frames in ADTS headers?
	if (audio->dwFlags & MKV_ADD_ADTS)
	{
		iADTSlen = 7;
	}
	else
	{
		iADTSlen = 0;
	}

	// Buffer size query?
	if (lpBuffer == NULL)
	{
		msi = &(audio->sample_list[dwStart]);
		actual = iADTSlen + audio->iCompSettingsLen + (msi->size & (~MKV_KEYFRAME));

		for (i = 1; i < lCount; ++i)
		{
			++msi;
			actual += iADTSlen + audio->iCompSettingsLen + (msi->size & (~MKV_KEYFRAME));
		}

		if (actual < audio->dwMaxSample) actual = audio->dwMaxSample;

		if (lBytesRead) *lBytesRead = actual;
		if (lSamplesRead) *lSamplesRead = i;
		return true;
	}

	// Buffer too small to hold a sample?
	if (cbBuffer < audio->dwMaxSample)
	{
		if (lBytesRead) *lBytesRead = audio->dwMaxSample;
		if (lSamplesRead) *lSamplesRead = 1;
		return false;
	}

	if (iADTSlen)
	{
		// Gather information for ADTS header	

		// AudioSpecificConfig:
		//   LSB      MSB
		// xxxxx... ........  object type (00010 = AAC LC)
		// .....xxx x.......  sampling rate
		// ........ .xxxx...  channel configuration
		// ........ .....x..  frame length flag (0 = 1024, 1 = 960)
		// ........ ......x.  depends on core coder (0)
		// ........ .......x  extension flag (0)

		profile = *((WORD *)(audio->pFormat + sizeof(WAVEFORMATEX)));
		profile = ((profile & 0xFF) << 8) | (profile >> 8);

		chcode  = (profile >> 3) & 7;
		srcode  = (profile >> 7) & 15;
		profile = ((profile >> 11) - 1) & 3;
	}

	// Begin reading samples
	dst = (unsigned char *)lpBuffer;
	msi = &(audio->sample_list[dwStart]);

	for (i = 0; i < lCount; ++i)
	{
		DWORD dwBlkSize  = msi->size & (~MKV_KEYFRAME);
		DWORD dwExpected = iADTSlen + audio->iCompSettingsLen + dwBlkSize;

		if (cbBuffer < dwExpected) break;

		if (iADTSlen)
		{
			// Build ADTS header:
			// 00: 11111111   syncword
			// 01: 1111....   syncword (continued)
			//     ....0...   id (1 = MPEG-2, 0 = MPEG-4)
			//     .....00.   layer
			//     .......1   protection absent
			// 02: xx......   profile
			//     ..xxxx..   sf_index
			//     ......0.   private bit
			//     .......x   channel configuration
			// 03: xx......   channel configuration (continued)
			//     ..0.....   original
			//     ...0....   home
			//     ....0...   copyright identification bit
			//     .....0..   copyright identification start
			//     ......xx   frame length
			// 04: xxxxxxxx   frame length (continued)
			// 05: xxx.....   frame length (continued)
			//     ...11111   buffer fullness
			// 06: 111111..   buffer fullness (continued)
			//     ......00   raw data blocks in frame

			unsigned int frlen = (7 + audio->iCompSettingsLen + dwBlkSize) & 0x1FFF;
			
			dst[0] = 0xFF;
			dst[1] = 0xF1;
			dst[2] = (unsigned char)((profile << 6) | (srcode << 2) | (chcode >> 2));
			dst[3] = (unsigned char)(((chcode & 3) << 6) | ((frlen >> 11) & 3));
			dst[4] = (unsigned char)((frlen >> 3) & 0xFF);
			dst[5] = (unsigned char)(((frlen & 7) << 5) | 0x1F);
			dst[6] = 0xFC;
		}

		if (audio->iCompSettingsLen > 0)
		{
			memcpy(dst + iADTSlen, audio->pCompSettings, audio->iCompSettingsLen);
		}

		actual = parentPtr->ReadData(dst + iADTSlen + audio->iCompSettingsLen, msi->pos, dwBlkSize);
		if (actual != dwBlkSize) break;
		
		dst      += dwExpected;
		cbBuffer -= dwExpected;

		++msi;
	}

	if (lBytesRead) *lBytesRead = (uint32)(dst - (unsigned char *)lpBuffer);
	if (lSamplesRead) *lSamplesRead = i;

	return true;
}


const void *AudioSourceMKV::GetDirectFormat()
{
	return audio->pFormat;
}


int AudioSourceMKV::GetDirectFormatLen()
{
	return audio->iFormatLen;
}


IVDXStreamSource::ErrorMode AudioSourceMKV::GetDecodeErrorMode()
{
	return errmode;
}


void AudioSourceMKV::SetDecodeErrorMode(IVDXStreamSource::ErrorMode mode)
{
	errmode = mode;
}


bool AudioSourceMKV::IsDecodeErrorModeSupported(IVDXStreamSource::ErrorMode mode)
{
	return true;
}


bool AudioSourceMKV::IsVBR()
{
	return (audio->dwFlags & MKV_IS_VBR)? true: false;
}


sint64 AudioSourceMKV::TimeToPositionVBR(sint64 us)
{
/*
	DWORD i;
	DWORD ms = (DWORD)((double)us / 1000.0);

	const MKVSampleInfo *msi = audio->sample_list;

	for (i = 0; i < audio->dwActualLength; ++i)
	{
		if (msi->pres_time >= ms) return (sint64)i;
		++msi;
	}

	return 0;
*/
	DWORD i;
	DWORD ms = (DWORD)((double)us / 1000.0 + 0.5);

	for (i = 1; i < audio->dwActualLength; ++i)
	{
		if (ms < audio->sample_list[i].pres_time)
		{
			if (us >= 0)
			{
				DWORD dwPrvDiff = ms - audio->sample_list[i - 1].pres_time;
				DWORD dwCurDiff = audio->sample_list[i].pres_time - ms;
			
				if (dwPrvDiff > dwCurDiff) ++i;
			}
			break;
		}
	}

	--i;
	return (sint64)i;
}


sint64 AudioSourceMKV::PositionToTimeVBR(sint64 samples)
{
/*
	DWORD i = (DWORD)samples;
	
	if (i < audio->dwActualLength)
	{
		return ((sint64)1000 * audio->sample_list[i].pres_time);
	}
	
	return 0;
*/
	DWORD i = (DWORD)samples;

	if (i < audio->dwActualLength)
	{
		return ((sint64)1000 * audio->sample_list[i].pres_time);
	}
	
	if (samples < 0) return 0;

	return (sint64)(audio->Duration * 1000.0 + 0.5);
}


void AudioSourceMKV::GetAudioSourceInfo(VDXAudioSourceInfo& info)
{
	info.mFlags = 0;
}


