///////////////////////////////////////////////////////////////////////
// Matroska plugin for VirtualDub 1.8.8+
// Copyright (C) 2010-2012 fccHandler
// Copyright (C) 1998-2012 Avery Lee
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////

// File Information dialog (part of InputFileMKV class)
// and other static utility functions

#if defined(_MSC_VER) && (_MSC_VER >= 1400)
	#define _CRT_SECURE_NO_WARNINGS
#endif

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <commctrl.h>
#include <vfw.h>
#include <ks.h>
#include <ksmedia.h>
#include <stdio.h>

#include "VDInputDriver.h"
#include "VDUnknown.h"

#include "MkvFile.h"
#include "InputOptionsMKV.h"
#include "DataVector.h"
#include "InputFileMKV.h"
#include "MatroskaTrack.h"
#include "Languages.h"
#include "Resource.h"


// static
int InputFileMKV::lang_to_id(const char *lang)
{
	if (lang != NULL)
	{
		const iso639_language_t *desc = iso639_languages;
		int i;

		for (i = 0; ; ++i, ++desc)
		{
			if (desc->english_name == NULL) break;

			if (desc->iso639_2_code != NULL)
			{
				if (_strnicmp(lang, desc->iso639_2_code, 3) == 0) return i;
			}
		}
	}
	return -1;
}


// static
bool InputFileMKV::IsPCM(const MatroskaTrack &mt)
{
	if (mt.iCodecID == A_PCM_INT_LIT) return true;
	if (mt.iCodecID == A_PCM_INT_BIG) return true;

	if (mt.iCodecID == A_MS_ACM && mt.iFormatLen >= sizeof(PCMWAVEFORMAT))
	{
		const WAVEFORMATPCMEX *wf = (WAVEFORMATPCMEX *)mt.pFormat;

		if (wf->Format.wFormatTag == WAVE_FORMAT_PCM) return true;

		if (mt.iFormatLen >= sizeof(WAVEFORMATPCMEX))
		{
			if (wf->Format.wFormatTag == WAVE_FORMAT_EXTENSIBLE)
			{
				if (wf->SubFormat == KSDATAFORMAT_SUBTYPE_PCM) return true;
			}
		}
	}

	return false;
}


// static
void InputFileMKV::SetItemTextW(HWND hDlg, int iItem, wchar_t *pText)
{
	if (pText != NULL)
	{
		if (IsWindowUnicode(hDlg) != FALSE)
		{
			SetDlgItemTextW(hDlg, iItem, pText);
			return;
		}
		else
		{
			char *pTextA = WinFile::WcharToAnsi(pText);
			if (pTextA != NULL)
			{
				SetDlgItemTextA(hDlg, iItem, pTextA);
				delete[] pTextA;
				return;
			}
		}
	}

	SetDlgItemTextA(hDlg, iItem, "");
}


// static
void InputFileMKV::SetDlgUserLong(HWND hDlg, LONG_PTR lParam)
{
	if (IsWindowUnicode(hDlg) != FALSE)
		SetWindowLongPtrW(hDlg, DWLP_USER, lParam);
	else
		SetWindowLongPtrA(hDlg, DWLP_USER, lParam);
}


// static
LONG_PTR InputFileMKV::GetDlgUserLong(HWND hDlg)
{
	if (IsWindowUnicode(hDlg) != FALSE)
		return GetWindowLongPtrW(hDlg, DWLP_USER);
	else
		return GetWindowLongPtrA(hDlg, DWLP_USER);
}


// static
char *InputFileMKV::time_to_string(double seconds)
{
	// There are 86400 seconds in a day,
	// about 32 million seconds in a year.
	// Since we are unlikely to encounter videos
	// of such an extreme duration, we can easily
	// get away with using ints for this math...

	const int buf_size = 32;
	char *ret = new char[buf_size];

	if (ret != NULL)
	{
		int hours, mins, mils;
		int secs = (int)seconds;

		hours = secs / 3600;
		secs  = secs % 3600;
		mins  = secs / 60;
		secs  = secs % 60;
		mils  = (int)(seconds * 1000.0 + 0.5);
		mils  = mils % 1000;

		if (hours > 0)
		{
			_snprintf(ret, buf_size, "%i:%02i:%02i.%03i", hours, mins, secs, mils);
		}
		else
		{
			_snprintf(ret, buf_size, "%i:%02i.%03i", mins, secs, mils);
		}

		ret[buf_size - 1] = '\0';
	}

	return ret;
}


typedef struct {
	__int64       i64VideoBytes;
	__int64       i64AudioBytes;
	__int64       i64FrameTotalSize[2];
	unsigned long dwFrameCount[2];
	unsigned long dwFrameMinSize[2];
	unsigned long dwFrameAvgSize[2];
	unsigned long dwFrameMaxSize[2];
} MKVStreamInfo;


// static
void InputFileMKV::UpdateDlgAudio(HWND hDlg, InputFileMKV *thisPtr, int index)
{
	// The longest signed int is "-2147483648" (11 chars)
	// The longest unsigned int is  "4294967295" (10 chars)
	// The longest signed int64 is "-9223372036854775808" (20 chars)

	char     buf[64];
	char*    s;
	int      i;
	bool     bIsPCM;

	unsigned long rate;

	MKVStreamInfo pInfo = { 0 };

	const WAVEFORMATEX  *wfex;
	const MatroskaTrack *mt;

	_snprintf(buf, sizeof(buf), "Stream %i of %i", index + 1, thisPtr->iAudioTracks);

	for (i = 0; i < thisPtr->iTracks; ++i)
	{
		if (thisPtr->Track[i]->iTrackType == MKV_TYPE_AUDIO)
		{
			if (index == 0)
			{
				mt = thisPtr->Track[i];
				break;
			}
			--index;
		}
	}
	if (i == thisPtr->iTracks)
	{
		SetDlgItemTextA(hDlg, IDC_STREAMNUM, "");
		return;
	}

	buf[sizeof(buf) - 1] = '\0';
	SetDlgItemTextA(hDlg, IDC_STREAMNUM, buf);

	wfex = (WAVEFORMATEX *)mt->pFormat;
	bIsPCM = IsPCM(*mt);

/********************************************************************
        Number of samples (duration)
 ********************************************************************/

	s = time_to_string(mt->Duration);
	_snprintf(buf, sizeof(buf), "%I64u samples (%s)",
		(unsigned __int64)(mt->Duration * wfex->nSamplesPerSec + 0.5), s);

	delete[] s;

	buf[sizeof(buf) - 1] = '\0';
	SetDlgItemTextA(hDlg, IDC_AUDIO_LENGTH, buf);

/********************************************************************
        Sampling rate, channels, bits per sample
 ********************************************************************/

	i = _snprintf(buf, sizeof(buf), "%lu Hz", wfex->nSamplesPerSec);

	if (mt->dwOutputSamplingFrequency != 0 &&
		mt->dwOutputSamplingFrequency != mt->dwSamplingFrequency)
	{
		i += _snprintf(buf + i, sizeof(buf) - i, " (SBR)");
	}

	if (wfex->nChannels == 1) {
		i += _snprintf(buf + i, sizeof(buf) - i, ", Mono");
	} else if (wfex->nChannels == 2) {
		i += _snprintf(buf + i, sizeof(buf) - i, ", Stereo");
	} else {
		i += _snprintf(buf + i, sizeof(buf) - i, ", %hu channels", wfex->nChannels);
	}

	if (bIsPCM && wfex->wBitsPerSample != 0)
	{
		_snprintf(buf + i, sizeof(buf) - i, ", %hu-bit", wfex->wBitsPerSample);
	} else {
		_snprintf(buf + i, sizeof(buf) - i, ", N/A");
	}

	buf[sizeof(buf) - 1] = '\0';
	SetDlgItemTextA(hDlg, IDC_AUDIO_SAMPLINGRATE, buf);

/********************************************************************
        ACM Codec, tag
 ********************************************************************/

	if (bIsPCM)
	{
		_snprintf(buf, sizeof(buf), "Uncompressed PCM (0x%04X)", wfex->wFormatTag);
	}
	else
	{
		// TODO: Support Unicode description
		ACMFORMATTAGDETAILS aftd = { 0 };

		aftd.cbStruct = sizeof(aftd);
		aftd.dwFormatTag = wfex->wFormatTag;

		i = acmFormatTagDetails(NULL, &aftd, ACM_FORMATTAGDETAILSF_FORMATTAG);

		if (i == MMSYSERR_NOERROR)
		{
			_snprintf(buf, sizeof(buf),
				"%s (0x%04X)", aftd.szFormatTag, wfex->wFormatTag);
		} else {
			_snprintf(buf, sizeof(buf), "Unknown (0x%04X)", wfex->wFormatTag);
		}
	}

	buf[sizeof(buf) - 1] = '\0';
	SetDlgItemTextA(hDlg, IDC_AUDIO_COMPRESSION, buf);

/********************************************************************
        Block layout
 ********************************************************************/

	_snprintf(buf, sizeof(buf), "%lu chunks", mt->dwActualLength);

	buf[sizeof(buf) - 1] = '\0';
	SetDlgItemTextA(hDlg, IDC_AUDIO_LAYOUT, buf);

/********************************************************************
        Min/avg/max(total) block size
 ********************************************************************/

	pInfo.i64AudioBytes     = 0;
	pInfo.dwFrameMinSize[0] = 0xFFFFFFFF;
	pInfo.dwFrameMaxSize[0] = 0;

	for (i = 0; (DWORD)i < mt->dwActualLength; ++i)
	{
		DWORD dwSize = (mt->sample_list[i].size & (~MKV_KEYFRAME)) + mt->iCompSettingsLen;

		pInfo.i64AudioBytes += dwSize;

		if (dwSize < pInfo.dwFrameMinSize[0])
			pInfo.dwFrameMinSize[0] = dwSize;

		if (dwSize > pInfo.dwFrameMaxSize[0])
			pInfo.dwFrameMaxSize[0] = dwSize;
	}

	// Adjust if we are adding ADTS headers
	if (mt->dwFlags & MKV_ADD_ADTS)
	{
		pInfo.dwFrameMinSize[0] += 7;
		pInfo.dwFrameMaxSize[0] += 7;
		pInfo.i64AudioBytes += 7 * mt->dwActualLength;
	}

	_snprintf(buf, sizeof(buf), "%lu / %lu / %lu (%I64i)",
		pInfo.dwFrameMinSize[0],
		(unsigned long)((double)pInfo.i64AudioBytes
		              / (double)mt->dwActualLength + 0.5),
		pInfo.dwFrameMaxSize[0],
		pInfo.i64AudioBytes
	);

	buf[sizeof(buf) - 1] = '\0';
	SetDlgItemTextA(hDlg, IDC_AUDIO_FRAMESIZES, buf);

/********************************************************************
        Data rate
 ********************************************************************/

	if (mt->Duration > 0.0) {
		rate = (unsigned long)((8.0 * pInfo.i64AudioBytes) / mt->Duration + 0.5);
	} else {
		rate = 0;
	}
	_snprintf(buf, sizeof(buf), "%lu bits/sec", rate);

	buf[sizeof(buf) - 1] = '\0';
	SetDlgItemTextA(hDlg, IDC_AUDIO_DATARATE, buf);

/********************************************************************
        Language (code)
 ********************************************************************/

	if (mt->iLanguageID >= 0)
	{
		_snprintf(buf, sizeof(buf) - 1, "%s (%s)",
			iso639_languages[mt->iLanguageID].english_name,
			iso639_languages[mt->iLanguageID].iso639_2_code
		);
		buf[sizeof(buf) - 1] = '\0';
	}
	else
	{
		_snprintf(buf, sizeof(buf), "English (unspecified)");
	}

	buf[sizeof(buf) - 1] = '\0';
	SetDlgItemTextA(hDlg, IDC_AUDIO_LANGUAGE, buf);
}


// static
inline void InputFileMKV::UpdateDlgVideo(HWND hDlg, InputFileMKV *thisPtr, int index)
{
	char buf[64];
	unsigned char fcc[5];
	int i;
	unsigned long rate;

	double fps;
	char *s;

	MKVStreamInfo pInfo = { 0 };

	const MKVSampleInfo    *msi;
	const BITMAPINFOHEADER *bmih;
	const MatroskaTrack    *mt;

	for (i = 0; i < thisPtr->iTracks; ++i)
	{
		if (thisPtr->Track[i]->iTrackType == MKV_TYPE_VIDEO)
		{
			if (index == 0)
			{
				mt = thisPtr->Track[i];
				break;
			}
			--index;
		}
	}
	if (i == thisPtr->iTracks) return;

	fps = (double)mt->dwScale;
	if (mt->dwScale) fps = (double)mt->dwRate / fps;

	bmih = (BITMAPINFOHEADER *)mt->pFormat;

/********************************************************************
        Frame size, average fps
 ********************************************************************/

	_snprintf(buf, sizeof(buf), "%ldx%ld, %.3f fps",
		bmih->biWidth, bmih->biHeight, fps);

	buf[sizeof(buf) - 1] = '\0';
	SetDlgItemTextA(hDlg, IDC_VIDEO_FORMAT, buf);

/********************************************************************
        Number of frames (duration)
 ********************************************************************/

	s = time_to_string((double)mt->dwActualLength / fps);
	_snprintf(buf, sizeof(buf), "%lu (%s)", mt->dwActualLength, s);
	delete[] s;

	buf[sizeof(buf) - 1] = '\0';
	SetDlgItemTextA(hDlg, IDC_VIDEO_NUMFRAMES, buf);

/********************************************************************
        VCM codec (fourcc)
 ********************************************************************/

	if (mt->hCodec != NULL)
	{
		// TODO: Support Unicode description
		ICINFO ici;

		ici.dwSize = sizeof(ici);
		if (ICGetInfo(mt->hCodec, &ici, sizeof(ici))
			>= FIELD_OFFSET(ICINFO, szDriver))
		{
			i = _snprintf(buf, sizeof(buf), "%ls", ici.szDescription);
		} else {
			i = _snprintf(buf, sizeof(buf), "Unknown");
		}

	} else {
		i = _snprintf(buf, sizeof(buf), "Missing codec");
	}

	*(DWORD *)fcc = bmih->biCompression;
	for (int n = 0; n < 4; ++n)
	{
		if (fcc[n] < 32) fcc[n] = '?';
	}
	fcc[4] = '\0';

	_snprintf(buf + i, sizeof(buf) - i, " (%s)", (char *)fcc);

	buf[sizeof(buf) - 1] = '\0';
	SetDlgItemTextA(hDlg, IDC_VIDEO_COMPRESSION, buf);

/********************************************************************
        Number of key frames
 ********************************************************************/

	for (i = 0; i < 2; ++i) {
		pInfo.dwFrameMinSize[i] = 0xFFFFFFFF;
	}

	for (i = 0; i < (int)mt->dwActualLength; ++i)
	{
		msi = &(mt->sample_list[i]);
		int iFrameType = (msi->size & MKV_KEYFRAME)? 0: 1;
		unsigned long dwSize = (msi->size & (~MKV_KEYFRAME)) + mt->iCompSettingsLen;

		pInfo.i64VideoBytes += dwSize;

		++pInfo.dwFrameCount[iFrameType];
		pInfo.i64FrameTotalSize[iFrameType] += dwSize;

		if (dwSize < pInfo.dwFrameMinSize[iFrameType])
			pInfo.dwFrameMinSize[iFrameType] = dwSize;

		if (dwSize > pInfo.dwFrameMaxSize[iFrameType])
			pInfo.dwFrameMaxSize[iFrameType] = dwSize;
	}

	_snprintf(buf, sizeof(buf), "%lu", pInfo.dwFrameCount[0]);

	buf[sizeof(buf) - 1] = '\0';
	SetDlgItemTextA(hDlg, IDC_VIDEO_NUMKEYFRAMES, buf);

/********************************************************************
        Min/avg/max(total) key frame size
 ********************************************************************/

	if (pInfo.dwFrameCount[0] != 0) {
		_snprintf(buf, sizeof(buf), "%lu / %lu / %lu (%I64i)",
			pInfo.dwFrameMinSize[0],
			(unsigned long)((double)pInfo.i64FrameTotalSize[0]
			              / (double)pInfo.dwFrameCount[0] + 0.5),
			pInfo.dwFrameMaxSize[0],
			pInfo.i64FrameTotalSize[0]
		);
	} else {
		_snprintf(buf, sizeof(buf), "(no key frames)");
	}

	buf[sizeof(buf) - 1] = '\0';
	SetDlgItemTextA(hDlg, IDC_VIDEO_KEYFRAMESIZES, buf);

/********************************************************************
        Min/avg/max(total) delta frame size
 ********************************************************************/

	if (pInfo.dwFrameCount[1] != 0) {
		_snprintf(buf, sizeof(buf), "%lu / %lu / %lu (%I64i)",
			pInfo.dwFrameMinSize[1],
			(unsigned long)((double)pInfo.i64FrameTotalSize[1]
			              / (double)pInfo.dwFrameCount[1] + 0.5),
			pInfo.dwFrameMaxSize[1],
			pInfo.i64FrameTotalSize[1]
		);
	} else {
		_snprintf(buf, sizeof(buf), "(no delta frames)");
	}

	buf[sizeof(buf) - 1] = '\0';
	SetDlgItemTextA(hDlg, IDC_VIDEO_NONKEYFRAMESIZES, buf);

/********************************************************************
        Data rate
 ********************************************************************/

	if (mt->Duration > 0.0) {
		rate = (unsigned long)((8.0 * pInfo.i64VideoBytes) / mt->Duration + 0.5);
	} else {
		rate = 0;
	}
	_snprintf(buf, sizeof(buf), "%lu bits/sec", rate);

	buf[sizeof(buf) - 1] = '\0';
	SetDlgItemTextA(hDlg, IDC_VIDEO_DATARATE, buf);
}


// static
INT_PTR APIENTRY InputFileMKV::InfoDlgProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	InputFileMKV *thisPtr = (InputFileMKV *)GetDlgUserLong(hDlg);

	switch (message) {

    case WM_INITDIALOG:
	{
		SetDlgUserLong(hDlg, lParam);
		thisPtr = (InputFileMKV *)lParam;

		InputOptionsMKV::InitDialogTitle(hDlg, "File Information");

		SetItemTextW(hDlg, IDC_MUXING_APP,  thisPtr->MuxingApp);
		SetItemTextW(hDlg, IDC_WRITING_APP, thisPtr->WritingApp);

		UpdateDlgVideo(hDlg, thisPtr, 0);

		if (thisPtr->iAudioTracks > 1)
		{
			HWND hwndSpin = GetDlgItem(hDlg, IDC_SPIN1);
			EnableWindow(hwndSpin, TRUE);
			SendMessageA(hwndSpin, UDM_SETRANGE, 0,
				(LPARAM)MAKELONG(thisPtr->iAudioTracks - 1, 0));
			SendMessageA(hwndSpin, UDM_SETPOS, 0, 0);
			ShowWindow(hwndSpin, SW_SHOW);
		}

		UpdateDlgAudio(hDlg, thisPtr, 0);

		return TRUE;
	}

	case WM_COMMAND:
		// Note: "IDCANCEL" is the [X] button
		// Return TRUE if we process the message

		if (LOWORD(wParam) == IDOK || LOWORD(wParam) == IDCANCEL)
		{
			EndDialog(hDlg, TRUE);
			return TRUE;
		}
		break;

	case WM_NOTIFY:
	{
		const NMUPDOWN *nmud = (NMUPDOWN *)lParam;

		if (nmud->hdr.idFrom == IDC_SPIN1 && nmud->hdr.code == UDN_DELTAPOS)
		{
			int iNewPos = nmud->iPos + nmud->iDelta;

			// The UDS_WRAP style does wrap the position,
			// but only after we get this message (aargh).
			// Therefore we have to wrap it manually...

			if (iNewPos < 0)
			{
				iNewPos = thisPtr->iAudioTracks - 1;
			}
			else if (iNewPos >= thisPtr->iAudioTracks)
			{
				iNewPos = 0;
			}

			UpdateDlgAudio(hDlg, thisPtr, iNewPos);
		}

		break;
	}

    }

    return FALSE;
}


