///////////////////////////////////////////////////////////////////////
// Matroska plugin for VirtualDub 1.8.8+
// Copyright (C) 2010-2012 fccHandler
// Copyright (C) 1998-2012 Avery Lee
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////

// InputFileMKV implementation

#if defined(_MSC_VER) && (_MSC_VER >= 1400)
	#define _CRT_SECURE_NO_WARNINGS
#endif

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <commctrl.h>
#include <vfw.h>
#include <stdio.h>

#include "VDInputDriver.h"
#include "VDUnknown.h"

#include "MkvFile.h"
#include "DataVector.h"
#include "InputFileMKV.h"
#include "InputOptionsMKV.h"
#include "VideoSourceMKV.h"
#include "AudioSourceMKV.h"
#include "MatroskaTrack.h"
#include "Resource.h"


HMODULE g_hModule;

BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
	if (fdwReason == DLL_PROCESS_ATTACH)
	{
		g_hModule = (HMODULE)hinstDLL;
	}
	return TRUE;
}


///////////////////////////////////////////////////////////////////////
//
//						InputFileMKV
//
///////////////////////////////////////////////////////////////////////

InputFileMKV::InputFileMKV(const VDXInputDriverContext& context)
	: mContext          (context)
	, MuxingApp         (NULL)
	, WritingApp        (NULL)
	, hwndProgress      (NULL)
	, fAbort            (false)
	, iTracks           (0)
	, iAudioTracks      (0)
	, iVideoTracks      (0)
	, iTimecodeScale    (1000000)	// nanoseconds
{
	for (int i = 0; i < MAX_TRACKS; ++i)
	{
		Track[i] = NULL;
	}
}


InputFileMKV::~InputFileMKV()
{
	int i;

	for (i = MAX_TRACKS - 1; i >= 0; --i)
	{
		delete Track[i];
	}

	delete[] WritingApp;
	delete[] MuxingApp;
}


int InputFileMKV::ReadData(void *lpBuffer, __int64 offset, int bytes)
{
	if (lpBuffer != NULL && mFile.Seek(offset))
	{
		return mFile.Read(lpBuffer, bytes);
	}
	return 0;
}


bool InputFileMKV::guiDlgMessageLoop(HWND hDlg)
{
	MSG msg;

	while(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
	{
		if (msg.message == WM_QUIT) return false;

		if (!hDlg || !IsWindow(hDlg) || !IsDialogMessage(hDlg, &msg))
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}

	return true;
}


BOOL CALLBACK InputFileMKV::FindContextWindow(HWND hwnd, LPARAM lParam)
{
	if (GetParent(hwnd) == NULL)
	{
		*(HWND *)lParam = hwnd;
		return FALSE;
	}
	return TRUE;
}


INT_PTR CALLBACK InputFileMKV::ParseDialogProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	InputFileMKV *thisPtr = (InputFileMKV *)GetDlgUserLong(hDlg);

	switch(uMsg) {

	case WM_INITDIALOG:
	{
		SetDlgUserLong(hDlg, lParam);
		thisPtr = (InputFileMKV *)lParam;

		SendMessage(GetDlgItem(hDlg, IDC_PROGRESS1), PBM_SETRANGE, 0, MAKELPARAM(0, 16384));
		SetTimer(hDlg, 1, 250, (TIMERPROC)NULL);
		ShowWindow(hDlg, SW_SHOW);
		return TRUE;
	}

	case WM_TIMER:
	{
		char buf[64];
		__int64 file_cpos = thisPtr->mFile.Pos();
		__int64 file_size = thisPtr->mFile.Size();
		
		SendMessage(GetDlgItem(hDlg, IDC_PROGRESS1), PBM_SETPOS,
			(WPARAM)((file_cpos * 16384I64) / file_size), 0);

		sprintf(buf, "%I64iK of %I64iK",
			file_cpos >> 10,
			(file_size + 512) >> 10);

		SendDlgItemMessage(hDlg, IDC_CURRENT_VALUE, WM_SETTEXT, 0, (LPARAM)buf);
		return TRUE;
	}

	case WM_COMMAND:
		if (LOWORD(wParam) == IDCANCEL) thisPtr->fAbort = true;
		return TRUE;

	case WM_DESTROY:
		KillTimer(hDlg, 1);
		break;

	default:
		break;
	}

	return FALSE;
}


bool InputFileMKV::ParseBlockGroup(__int64 limit, const __int64 timebase)
{
	// Parse one Matroska BlockGroup

	int id;
	__int64 size;
	__int64 origin;

	// Aargh, we need to parse all state before adding the Block!

	// The only way to know whether a frame is a keyframe is to
	// look for the "ReferenceBlock" entry.  If present, then
	// this frame is not a keyframe, otherwise it is.  But we
	// need to parse the entire block before we can know this
	// because it may appear AFTER the block data, and often
	// does.

	// Block state
	int     BlockPending;
	int     Flags;
	bool    Keyframe;
	
	MKVSampleInfo msi;

	BlockPending = 0;
	Flags = 0;
	Keyframe = true;

	// Clamp "limit" to prevent an endless loop
	if (limit > mFile.Size()) limit = mFile.Size();

	while (mFile.Pos() < limit)
	{
		if (hwndProgress != NULL)
		{
			// Update the progress dialog
			guiDlgMessageLoop(hwndProgress);
			if (fAbort) break;
		}

		id = mFile.ElementID();
		size = mFile.ElementSize();
		origin = mFile.Pos();
		
		switch (id) {

		case 0x8E:			// Slices
		case 0x9B:			// BlockDuration
		case 0xA0:          // (unknown)
			break;

		case 0xA1:			// Block
		{
			__int64 tc;
			int tn;
			int i;

			if (BlockPending)
			{
				if (Keyframe) Flags |= 0x80000000;
				if (!Track[BlockPending - 1]->AddBlock(mFile, msi, Flags)) goto Abort;
				BlockPending = 0;
				Keyframe = true;
			}

			// TrackNumber
			tn = (int)mFile.MatroskaUint();

			// Timecode (relative to Cluster)
			tc = mFile.MotorolaInt(2);
			if (tc > 32767i64) tc -= 65536i64;
			tc = (timebase + tc) * iTimecodeScale;

			// Convert to milliseconds (from nanoseconds)
			msi.pres_time = (unsigned long)
				(((double)tc / 1000000.0) + 0.5);

			Flags = mFile.ReadByte();

			msi.pos = mFile.Pos();
			msi.size = (unsigned long)((origin + size) - msi.pos);

			for (i = 0; i < iTracks; ++i)
			{
				if (Track[i]->iTrackNumber == tn)
				{
					BlockPending = i + 1;
					break;
				}
			}

			break;
		}

		case 0xA2:			// BlockVirtual
		case 0xA4:			// CodecState
		case 0xBF:			// CRC-32
		case 0xEC:			// Void
		case 0xFA:			// ReferencePriority
			break;

		case 0xFB:			// ReferenceBlock
			Keyframe = false;
			break;

		case 0xFD:			// ReferenceVirtual
		case 0x75A1:		// BlockAdditions
		default:
			break;

		}

		mFile.Seek(origin + size);
	}

	if (BlockPending)
	{
		if (Keyframe) Flags |= 0x80000000;
		if (!Track[BlockPending - 1]->AddBlock(mFile, msi, Flags)) goto Abort;
	}

	return true;

Abort:
	return false;
}


bool InputFileMKV::ParseCluster(__int64 limit)
{
	// Parse one Matroska Cluster

	int id;
	__int64 size;
	__int64 origin;

	__int64 iTimecode = 0;

	// Clamp "limit" to prevent an endless loop
	if (limit > mFile.Size()) limit = mFile.Size();

	while (mFile.Pos() < limit)
	{
		id = mFile.ElementID();
		size = mFile.ElementSize();
		origin = mFile.Pos();

		switch (id) {

		case 0xA0:			// BlockGroup
			if (!ParseBlockGroup(origin + size, iTimecode)) goto Abort;
			break;

		case 0xA3:			// SimpleBlock
		{
			MKVSampleInfo msi;
			__int64 tc;
			int Flags;
			int i;

			if (hwndProgress != NULL)
			{
				// Update the progress dialog
				guiDlgMessageLoop(hwndProgress);
				if (fAbort) break;
			}

			// TrackNumber
			int tn = (int)mFile.MatroskaUint();

			// Timecode (relative to Cluster)
			tc = mFile.MotorolaInt(2);
			if (tc > 32767i64) tc -= 65536i64;
			tc = (iTimecode + tc) * iTimecodeScale;

			// Convert to milliseconds (from nanoseconds)
			msi.pres_time = (unsigned long)
				(((double)tc / 1000000.0) + 0.5);

			Flags = mFile.ReadByte();
			msi.pos = mFile.Pos();
			msi.size = (unsigned long)((origin + size) - msi.pos);

			if (Flags & 0x80) Flags |= 0x80000000;

			for (i = 0; i < iTracks; ++i)
			{
				if (Track[i]->iTrackNumber == tn)
				{
					if (!Track[i]->AddBlock(mFile, msi, Flags)) goto Abort;
					break;
				}
			}
			break;
		}

		case 0xA7:			// Position
		case 0xAB:			// PrevSize
		case 0xAF:			// EncryptedBlock
		case 0xBF:			// CRC-32
			break;

		case 0xE7:			// Timecode
			iTimecode = mFile.MotorolaInt((int)size);
			break;

		case 0xEC:			// Void
		case 0x5854:		// SilentTracks
		default:
			break;

		}

		if (fAbort) break;
		
		mFile.Seek(origin + size);
	}

	return true;

Abort:
	return false;
}


bool InputFileMKV::ParseTrackEntry(__int64 limit)
{
	// Parse one Matroska TrackEntry

	if (iTracks >= MAX_TRACKS)
	{
		mContext.mpCallbacks->SetError("Too many tracks!");
		goto Abort;
	}

	Track[iTracks] = new MatroskaTrack;
	if (Track[iTracks] == NULL)
	{
		mContext.mpCallbacks->SetErrorOutOfMemory();
		goto Abort;
	}

	if (Track[iTracks]->ParseTrackEntry(mFile, limit))
	{
		++iTracks;
	}
	else
	{
		// Track parse failed, but it could just
		// be an unsupported audio codec. Discard
		// the track, but return true anyway.

		delete Track[iTracks];
		Track[iTracks] = NULL;
	}

	return true;

Abort:
	return false;
}


bool InputFileMKV::ParseTracks(__int64 limit)
{
	// Parse Matroska Tracks list

	int id;
	__int64 size;
	__int64 origin;

	// Clamp "limit" to prevent an endless loop
	if (limit > mFile.Size()) limit = mFile.Size();

	while (mFile.Pos() < limit)
	{
		id = mFile.ElementID();
		size = mFile.ElementSize();
		origin = mFile.Pos();

		switch (id) {

		case 0xAE:			// TrackEntry
			if (!ParseTrackEntry(origin + size)) goto Abort;
			break;

		case 0xBF:			// CRC-32
		case 0xE2:			// TrackOperation
		case 0xEC:			// Void
		default:
			break;

		}

		mFile.Seek(origin + size);
	}

	return true;

Abort:
	return false;
}


void InputFileMKV::ParseInfo(__int64 limit)
{
	// Parse Matroska Segment Info

	int id;
	__int64 size;
	__int64 origin;

	// Clamp "limit" to prevent an endless loop
	if (limit > mFile.Size()) limit = mFile.Size();

	while (mFile.Pos() < limit)
	{
		id = mFile.ElementID();
		size = mFile.ElementSize();
		origin = mFile.Pos();

		switch (id) {

		case 0xBF:			// CRC-32
		case 0xEC:			// Void
		case 0x4444:		// SegmentFamily
		case 0x4461:		// DateUTC
		case 0x4489:		// Duration
			break;

		case 0x4D80:		// MuxingApp
		case 0x80:          // MuxingApp (libmkv 0.6.4 + HandBrake 0.9.4) 
			MuxingApp = mFile.MatroskaStringW((int)size);
			break;

		case 0x5741:		// WritingApp
			WritingApp = mFile.MatroskaStringW((int)size);
			break;

		case 0x6924:		// ChapterTranslate
		case 0x7384:		// SegmentFilename
		case 0x73A4:		// SegmentUID
		case 0x7BA9:		// Title
			break;

		case 0x2AD7B1:		// TimecodeScale
			iTimecodeScale = mFile.MotorolaInt((int)size);
			break;

		case 0x3C83AB:		// PrevFilename
		case 0x3CB923:		// PrevUID
		case 0x3E83BB:		// NextFilename
		case 0x3EB923:		// NextUID
		default:
			break;

		}

		mFile.Seek(origin + size);
	}
}


bool InputFileMKV::ParseSegment(__int64 limit)
{
	// Parse one Matroska Segment

	int id;
	__int64 size;
	__int64 resync;
	__int64 origin;

	// Clamp "limit" to prevent an endless loop
	if (limit > mFile.Size()) limit = mFile.Size();

	while (mFile.Pos() < limit)
	{
		resync = mFile.Pos();
		id = mFile.ElementID();
		size = mFile.ElementSize();
		origin = mFile.Pos();

		switch (id) {

		case 0xBF:			// CRC-32
		case 0xEC:			// Void
		case 0x1043A770:	// Chapters
		case 0x114D9B74:	// SeekHead
		case 0x1254C367:	// Tags
			break;

		case 0x1549A966:	// Info
			ParseInfo(origin + size);
			break;

		case 0x1654AE6B:	// Tracks
			if (!ParseTracks(origin + size)) goto Abort;
			break;

		case 0x1941A469:	// Attachments
		case 0x1C53BB6B:	// Cues
			break;

		case 0x1F43B675:	// Cluster
			if (!ParseCluster(origin + size))
			{
				mContext.mpCallbacks->SetError("ParseCluster failed.");
				goto Abort;
			}
			break;

		default:
			// Unknown Element ID, try to resync.
			origin = resync;
			size   = 1;
			break;

		}

		if (fAbort) break;

		mFile.Seek(origin + size);
	}

	return true;

Abort:
	return false;
}


bool InputFileMKV::ParseHeader(__int64 limit)
{
	// Parse Matroska Header

	int id;
	__int64 size;
	__int64 origin;

	// Clamp "limit" to prevent an endless loop
	if (limit > mFile.Size()) limit = mFile.Size();

	while (mFile.Pos() < limit)
	{
		id = mFile.ElementID();
		size = mFile.ElementSize();
		origin = mFile.Pos();

		switch (id) {

		case 0xBF:			// CRC-32
		case 0xEC:			// Void
			break;

		case 0x4282:		// DocType
		{
			int cmp;

			char *DocType = mFile.MatroskaStringA((int)size);
			if (DocType == NULL) goto Abort;

			cmp = strcmp(DocType, "matroska");
			if (cmp) cmp = strcmp(DocType, "webm");

			delete[] DocType;
			if (cmp) goto Abort;

			break;
		}

		case 0x4285:		// DocTypeReadVersion
		{
			int ver = mFile.MotorolaInt((int)size);
			if (ver != 1 && ver != 2) goto Abort;
			break;
		}

		case 0x4286:		// EBMLVersion
		case 0x4287:		// DocTypeVersion
		case 0x42F2:		// EBMLMaxIDLength
		case 0x42F3:		// EBMLMaxSizeLength
			break;

		case 0x42F7:		// EBMLReadVersion
			if (mFile.MotorolaInt((int)size) != 1) goto Abort;
			break;

		default:
			break;

		}

		mFile.Seek(origin + size);
	}

	return true;

Abort:
	return false;
}


bool InputFileMKV::ParseRoot()
{
	HWND hwndParent;

	int id;
	__int64 size;
	__int64 origin;

	bool bFoundHeader = false;
	bool ret = false;

	fAbort = false;

	// Get a handle to the main VirtualDub window
	hwndParent = NULL;
	EnumThreadWindows(GetCurrentThreadId(), FindContextWindow, (LPARAM)&hwndParent);
	if (hwndParent == NULL) hwndParent = GetActiveWindow();
	if (hwndParent != NULL) EnableWindow(hwndParent, FALSE);

	// Pop up parse dialog
	hwndProgress = CreateDialogParamA(g_hModule,
		MAKEINTRESOURCE(IDD_PROGRESS), hwndParent,
		ParseDialogProc, (LPARAM)this);

	while (!mFile.EndOfFile())
	{
		id = mFile.ElementID();
		size = mFile.ElementSize();
		origin = mFile.Pos();

		if (size < 0 || size > (mFile.Size() - origin))
		{
			size = mFile.Size() - origin;
		}

		switch (id) {

		case 0xBF:			// CRC-32
		case 0xEC:			// Void
			break;

		case 0x18538067:	// Segment
			if (!bFoundHeader)
			{
				mContext.mpCallbacks->SetError("Segment found without Header!");
				goto Abort;
			}
			if (!ParseSegment(origin + size)) goto Abort;
			break;

		case 0x1A45DFA3:	// Header
			if (!ParseHeader(origin + size))
			{
				mContext.mpCallbacks->SetError("ParseHeader failed.");
				goto Abort;
			}
			bFoundHeader = true;
			break;

		default:
			break;

		}

		if (fAbort) break;

		mFile.Seek(origin + size);
	}

	ret = true;

Abort:
	if (hwndParent != NULL) EnableWindow(hwndParent, TRUE);

	if (hwndProgress != NULL)
	{
		DestroyWindow(hwndProgress);
		hwndProgress = NULL;
	}

	return ret;
}


const wchar_t *InputFileMKV::GetFileTitle(const wchar_t *szFile)
{
	// Return a pointer to the filename minus the path
	int i;
	int sep = 0;

	for (i = 0; ; ++i)
	{
		const wchar_t ch = szFile[i];
		if (ch == L'\0') break;
		if (ch == L'\\' || ch == L'/') sep = i + 1;
	}

	return (szFile + sep);
}


void InputFileMKV::Init(const wchar_t *szFile, IVDXInputOptions *opts)
{
	struct InputOptionsMKV::OptionsMKV tOptions;

	bool   fHasIndex;
	bool   result;

	if (!mFile.Open(szFile))
	{
		mContext.mpCallbacks->SetError("Couldn't open %ls", GetFileTitle(szFile));
		return;
	}

	if (opts != NULL)
	{
		// Get options from VirtualDub
		InputOptionsMKV *pOptions = (InputOptionsMKV *)opts;
		pOptions->Write(&tOptions, sizeof(tOptions));
	}
	else
	{
		// Get options (if any) from the registry
		InputOptionsMKV pOptions;
		pOptions.Write(&tOptions, sizeof(tOptions));
	}

	fHasIndex = ReadIndex(szFile);
	if (fHasIndex)
	{
		// Did the user open the index file?
		DWORD ver[2];
		mFile.Read(ver, sizeof(ver));
		if (ver[0] == 'DIKM' && (ver[1] & 0xF8FFFFFF) == '000X')
		{
			// Yep, replace with Matroska file
			wchar_t *szMkv = GetFileNameFromIndex(szFile);
			if (!mFile.Open(szMkv))
			{
				mContext.mpCallbacks->SetError("Couldn't find %ls", GetFileTitle(szMkv));
				delete[] szMkv;
				return;
			}
			delete[] szMkv;
		}
	}
	else
	{
		// No index; do a full parse.
		int i;
		result = ParseRoot();

		if (!result) return;

		i = 0;
		while (i < iTracks)
		{
			// Deliver raw AAC?
			if (Track[i]->iCodecID == A_AAC)
			{
				if (tOptions.fDeliverRawAAC)
					Track[i]->dwFlags &= ~MKV_ADD_ADTS;
				else
					Track[i]->dwFlags |=  MKV_ADD_ADTS;
			}

			if (!Track[i]->Finish(mFile))
			{
				// Bad track; remove it
				int j;

				delete Track[i];

				for (j = i + 1; j < iTracks; ++j)
				{
					Track[j - 1] = Track[j];
				}

				Track[--iTracks] = NULL;
			}
			else
			{
				// Use the 'H264' FOURCC instead of 'AVC1'?
				if (tOptions.fUseH264NotAVC1 && Track[i]->iTrackType == MKV_TYPE_VIDEO)
				{
					BITMAPINFOHEADER *bmih = (BITMAPINFOHEADER *)Track[i]->pFormat;
					if (bmih->biCompression == '1CVA') bmih->biCompression = '462H';
				}
				// OK; goto next track
				++i;
			}
		}

		for (i = 0; i < iTracks; ++i)
		{
			if (Track[i]->iTrackType == MKV_TYPE_AUDIO) ++iAudioTracks;
			if (Track[i]->iTrackType == MKV_TYPE_VIDEO) ++iVideoTracks;
		}
	}

	// Write an index?
	if (!fHasIndex && tOptions.fCreateIndex)
	{
		if (!WriteIndex(szFile))
		{
			// Failed writing index; delete it
			wchar_t *szIndex = GetIndexName(szFile);
			if (szIndex != NULL)
			{
				WinFile::Delete(szIndex);
				delete[] szIndex;
			}
		}
	}
}


bool InputFileMKV::Append(const wchar_t *szFile)
{
	return false;
}


bool InputFileMKV::PromptForOptions(VDXHWND hwndParent, IVDXInputOptions **ppOptions)
{
	InputOptionsMKV *p = new InputOptionsMKV;

	*ppOptions = p;
	if (p == NULL) return false;

	p->AddRef();

	p->OptionsDlg((HWND)hwndParent);
	return true;
}


bool InputFileMKV::CreateOptions(const void *buf, uint32 len, IVDXInputOptions **ppOptions)
{
	InputOptionsMKV *p = new InputOptionsMKV;

	*ppOptions = p;
	if (p == NULL) return false;

	p->AddRef();

	if (len > p->opts.len) len = p->opts.len;
	memcpy(&(p->opts), buf, len);

	return true;
}


void InputFileMKV::DisplayInfo(VDXHWND hwndParent)
{
	if (IsWindowUnicode((HWND)hwndParent) != FALSE)
	{
		DialogBoxParamW(g_hModule, MAKEINTRESOURCEW(IDD_MKV_INFO),
			(HWND)hwndParent, InfoDlgProc, (LPARAM)this);
	}
	else
	{
		DialogBoxParamA(g_hModule, MAKEINTRESOURCEA(IDD_MKV_INFO),
			(HWND)hwndParent, InfoDlgProc, (LPARAM)this);
	}
}


bool InputFileMKV::GetVideoSource(int index, IVDXVideoSource **ppVS)
{
	int i;
	*ppVS = NULL;

	for (i = 0; i < iTracks; ++i)
	{
		if (Track[i]->iTrackType == MKV_TYPE_VIDEO)
		{
			if (index == 0)
			{
				*ppVS = new VideoSourceMKV(this, Track[i], NULL);
				if (*ppVS == NULL) break;

				(*ppVS)->AddRef();
				return true;
			}
			--index;
		}
	}

	// No video track; choose the longest audio track
	if (index == 0)
	{
		double max = 0.0;
		index = -1;

		for (i = iTracks - 1; i >= 0; --i)
		{
			if (Track[i]->iTrackType == MKV_TYPE_AUDIO)
			{
				if (Track[i]->Duration >= max)
				{
					max   = Track[i]->Duration;
					index = i;
				}
			}
		}

		if (index >= 0)
		{
			*ppVS = new VideoSourceMKV(this, NULL, Track[index]);
			if (*ppVS != NULL)
			{
				(*ppVS)->AddRef();
				return true;
			}
		}
	}

	return false;
}


bool InputFileMKV::GetAudioSource(int index, IVDXAudioSource **ppAS)
{
	int i;
	*ppAS = NULL;

	for (i = 0; i < iTracks; ++i)
	{
		if (Track[i]->iTrackType == MKV_TYPE_AUDIO)
		{
			if (index == 0)
			{
				*ppAS = new AudioSourceMKV(this, Track[i]);
				if (*ppAS == NULL) break;

				(*ppAS)->AddRef();
				return true;
			}
			--index;
		}
	}

	return false;
}


///////////////////////////////////////////////////////////////////////
//
//						InputFileDriverMKV
//
///////////////////////////////////////////////////////////////////////

class InputFileDriverMKV : public vdxunknown<IVDXInputFileDriver> {
public:
	InputFileDriverMKV(const VDXInputDriverContext& context);
	~InputFileDriverMKV();

	int		VDXAPIENTRY DetectBySignature(const void *pHeader, sint32 nHeaderSize, const void *pFooter, sint32 nFooterSize, sint64 nFileSize);
	bool	VDXAPIENTRY CreateInputFile(uint32 flags, IVDXInputFile **ppFile);

protected:
	const VDXInputDriverContext& mContext;
};

InputFileDriverMKV::InputFileDriverMKV(const VDXInputDriverContext& context)
	: mContext(context)
{
}

InputFileDriverMKV::~InputFileDriverMKV() {
}

int VDXAPIENTRY InputFileDriverMKV::DetectBySignature(const void *pHeader, sint32 nHeaderSize, const void *pFooter, sint32 nFooterSize, sint64 nFileSize) {
	return -1;
}

bool VDXAPIENTRY InputFileDriverMKV::CreateInputFile(uint32 flags, IVDXInputFile **ppFile) {
	IVDXInputFile *p = new InputFileMKV(mContext);
	if (!p) return false;

	*ppFile = p;
	p->AddRef();
	return true;
}

bool VDXAPIENTRY MKV_create(const VDXInputDriverContext *pContext, IVDXInputFileDriver **ppDriver) {
	IVDXInputFileDriver *p = new InputFileDriverMKV(*pContext);
	if (!p) return false;

	*ppDriver = p;
	p->AddRef();
	return true;
}

const uint8 MKV_sig[] = {
	0x1A, 0xFF,
	0x45, 0xFF,
	0xDF, 0xFF,
	0xA3, 0xFF
};

const uint8 mkidx_sig[] = {
	'M', 0xFF,
	'K', 0xFF,
	'I', 0xFF,
	'D', 0xFF,
	'X', 0xFF,
	'0', 0xFF,
	'0', 0xFF,
	'0', 0xF8	// '1' (0x31), '2' (0x32), '3' (0x33), or '4' (0x34)
};

const VDXInputDriverDefinition MKV_input = {
	sizeof(VDXInputDriverDefinition),
	VDXInputDriverDefinition::kFlagSupportsVideo |
		VDXInputDriverDefinition::kFlagSupportsAudio,
	0,
	sizeof(MKV_sig),
	MKV_sig,
	L"*.mkv|*.mka",
	L"Matroska files (*.mkv,*.mka)|*.mkv;*.mka",
	L"MKV",
	MKV_create
};

const VDXInputDriverDefinition mkidx_input = {
	sizeof(VDXInputDriverDefinition),       // size of this structure
	VDXInputDriverDefinition::kFlagSupportsVideo |
		VDXInputDriverDefinition::kFlagSupportsAudio,
	0,                                      // priority
	sizeof(mkidx_sig),                      // size of file signature
	mkidx_sig,                              // pointer to signature
	L"*.mkidx",                             // extension pattern
	L"Matroska index (*.mkidx)|*.mkidx",
	L"MKIDX",                               // driver target name
	MKV_create                              // create proc
};

const VDXPluginInfo MKV_info = {
	sizeof(VDXPluginInfo),              // size
	L"Matroska plugin",                 // name
	L"fccHandler",                      // author
	L"Loads Matroska files.",           // description
	0x03010000,                         // (major<<24) + (minor<<16) + build
	kVDXPluginType_Input,               // type
	0,                                  // flags
	10,									// version required
	kVDXPlugin_APIVersion,              // version used
	2,									// version required
	kVDXPlugin_InputDriverAPIVersion,   // version used
	&MKV_input
};

const VDXPluginInfo mkidx_info = {
	sizeof(VDXPluginInfo),              // size of this structure
	L"MKIDX",                           // name of this plugin
	L"fccHandler",                      // author's name
	L"Loads Matroska index files.",     // plugin description
	0x04000000,                         // version = 4.0.0
	kVDXPluginType_Input,               // type of plugin = input
	0,                                  // flags = 0
	10,                                 // version required
	kVDXPlugin_APIVersion,              // version used
	2,                                  // version required
	kVDXPlugin_InputDriverAPIVersion,   // version used
	&mkidx_input                        // driver definition
};

///////////////////////////////////////////////////////////////////////////

const VDXPluginInfo *const kPlugins[] = {
	&MKV_info,
	&mkidx_info,
	NULL
};

extern "C" const VDXPluginInfo *const * VDXAPIENTRY VDGetPluginInfo() {
	return kPlugins;
}
