///////////////////////////////////////////////////////////////////////
// Matroska plugin for VirtualDub 1.8.8+
// Copyright (C) 2010-2012 fccHandler
// Copyright (C) 1998-2012 Avery Lee
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////

// MatroskaTrack implementation

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <vfw.h>

#ifdef _DEBUG
#include <stdio.h>
#endif

#include "VDInputDriver.h"
#include "VDUnknown.h"
#include "MkvFile.h"

#include "DataVector.h"

#include "InputFileMKV.h"
#include "MatroskaTrack.h"


// Invalid timestamp value
#define PTS_INVALID 0xFFFFFFFF


MatroskaTrack::MatroskaTrack()
{
	dv                        = NULL;

	wBitsPerSample            = 0;
	nChannels                 = 0;

	iPixelWidth               = 0;
	iPixelHeight              = 0;

	hCodec                    = NULL;
	iCodecID                  = 0;
	iLanguageID               = -1;

	pFormat                   = NULL;
	iFormatLen                = 0;

	dwSamplingFrequency       = 0;
	dwOutputSamplingFrequency = 0;

	pCodecPrivate             = NULL;
	iCodecPrivateLen          = 0;
	pCompSettings             = NULL;
	iCompSettingsLen          = 0;
	avc_len                   = 0;

	sample_list               = NULL;
	dwActualLength            = 0;
	ui64TotalBytes            = 0;
	dwMaxSample               = 0;
	Duration                  = 0.0;

	dwRate                    = 1;
	dwScale                   = 1;
	dwFlags                   = 0;

	iTrackNumber              = 0;
	iTrackType                = 0;
}


MatroskaTrack::~MatroskaTrack()
{
	delete[] sample_list;
	delete[] pFormat;
	delete[] pCompSettings;
	delete[] pCodecPrivate;
	delete dv;
}


bool MatroskaTrack::AddBlock(MkvFile &mFile, MKVSampleInfo &msi, int Flags)
{
	bool ret = false;

	// Flags is:
	// x....... ........ ........ ........  keyframe
	// .xxxxxxx xxxxxxxx xxxxxxxx xxxx....  reserved
	// ........ ........ ........ ....x...  invisible
	// ........ ........ ........ .....xx.  lacing
	// ........ ........ ........ .......x  unused

	if (Flags & 6)
	{
		// Aargh, laced Block...
		unsigned long *Sizes;
		int LacedFrames;
		int i;

		if (dv == NULL)
		{
			dv = new DataVector(sizeof(MKVSampleInfo));
			if (dv == NULL) goto Abort;
		}

		mFile.Seek(msi.pos);
		LacedFrames = mFile.ReadByte() + 1;

		Sizes = new unsigned long[LacedFrames];
		if (Sizes == NULL) goto Abort;

		if ((Flags & 6) == 2)
		{
			// Xiph lacing

			unsigned long total = msi.size - 1;
			unsigned long sz;
			int byte;

			for (i = 0; i < (LacedFrames - 1); ++i)
			{
				sz = 0;
				do {
					byte = mFile.ReadByte();
					--total;
					sz += byte;
				} while (byte == 255);

				Sizes[i] = sz;
				total -= sz;
			}
			Sizes[i] = total;

			msi.pos = mFile.Pos();
			msi.size = Sizes[0];
			if (Flags & 0x80000000) msi.size |= MKV_KEYFRAME;
			dv->Add(&msi);
			msi.size &= (~MKV_KEYFRAME);
			msi.pres_time = PTS_INVALID;

			for (i = 1; i < LacedFrames; ++i)
			{
				msi.pos += msi.size;
				msi.size = Sizes[i];
				dv->Add(&msi);
			}
		}
		else if ((Flags & 6) == 4)
		{
			// Fixed-size lacing
			msi.pos = mFile.Pos();
			msi.size = (msi.size - 1) / LacedFrames;

			if (Flags & 0x80000000) msi.size |= MKV_KEYFRAME;
			dv->Add(&msi);
			msi.size &= (~MKV_KEYFRAME);
			msi.pres_time = PTS_INVALID;

			for (i = 1; i < LacedFrames; ++i)
			{
				msi.pos += msi.size;
				dv->Add(&msi);
			}
		}
		else	// (flags & 6) == 6
		{
			// EBML lacing

			unsigned long total = msi.size - 1;
			__int64 origin2 = mFile.Pos();

			Sizes[0] = (int)mFile.MatroskaUint();
			total -= Sizes[0];

			for (i = 1; i < (LacedFrames - 1); ++i)
			{
				Sizes[i] = Sizes[i - 1] + (int)mFile.MatroskaSint();
				total -= Sizes[i];
			}
			total -= (unsigned long)(mFile.Pos() - origin2);
			Sizes[i] = total;

			msi.pos = mFile.Pos();
			msi.size = Sizes[0];
			if (Flags & 0x80000000) msi.size |= MKV_KEYFRAME;
			dv->Add(&msi);
			msi.size &= (~MKV_KEYFRAME);
			msi.pres_time = PTS_INVALID;

			for (i = 1; i < LacedFrames; ++i)
			{
				msi.pos += msi.size;
				msi.size = Sizes[i];
				dv->Add(&msi);
			}
		}

		delete[] Sizes;
		ret = true;
	}
	else
	{
		// No lacing, thank goodness
		if (dv == NULL)
		{
			dv = new DataVector(sizeof(MKVSampleInfo));
			if (dv == NULL) goto Abort;
		}

		if (Flags & 0x80000000) msi.size |= MKV_KEYFRAME;
		dv->Add(&msi);
		ret = true;
	}

Abort:
	return ret;
}


inline bool MatroskaTrack::ParseContentCompression(MkvFile &mFile, __int64 limit)
{
	int id;
	__int64 size;
	__int64 origin;

	// Clamp "limit" to prevent an endless loop
	if (limit > mFile.Size()) limit = mFile.Size();

	while (mFile.Pos() < limit)
	{
		id = mFile.ElementID();
		size = mFile.ElementSize();
		origin = mFile.Pos();

		switch (id) {

		case 0x4254:    // ContentCompAlgo
			if (mFile.MotorolaInt((int)size) != 3) goto Abort;
			break;

		case 0x4255:    // ContentCompSettings
			delete[] pCompSettings;
			iCompSettingsLen = 0;

			pCompSettings = new unsigned char[(size_t)size];
			if (pCompSettings == NULL) goto Abort;
			iCompSettingsLen = (int)size;

			mFile.Read(pCompSettings, iCompSettingsLen);
			break;

		case 0x5035:    // ContentEncryption
		default:
			break;

		}

		mFile.Seek(origin + size);
	}

	return true;

Abort:
	return false;
}


inline bool MatroskaTrack::ParseContentEncoding(MkvFile &mFile, __int64 limit)
{
	int id;
	__int64 size;
	__int64 origin;

	// Clamp "limit" to prevent an endless loop
	if (limit > mFile.Size()) limit = mFile.Size();

	while (mFile.Pos() < limit)
	{
		id = mFile.ElementID();
		size = mFile.ElementSize();
		origin = mFile.Pos();

		switch (id) {

		case 0x5031:    // ContentEncodingOrder
		case 0x5032:    // ContentEncodingScope
		case 0x5033:    // ContentEncodingType
			break;

		case 0x5034:    // ContentCompression
			if (!ParseContentCompression(mFile, origin + size)) goto Abort;
			break;

		case 0x5035:    // ContentEncryption
		default:
			break;

		}

		mFile.Seek(origin + size);
	}

	return true;

Abort:
	return false;
}


inline bool MatroskaTrack::ParseContentEncodings(MkvFile &mFile, __int64 limit)
{
	int id;
	__int64 size;
	__int64 origin;

	// Clamp "limit" to prevent an endless loop
	if (limit > mFile.Size()) limit = mFile.Size();

	while (mFile.Pos() < limit)
	{
		id = mFile.ElementID();
		size = mFile.ElementSize();
		origin = mFile.Pos();

		switch (id) {

		case 0x6240:    // ContentEncoding
			if (!ParseContentEncoding(mFile, origin + size)) goto Abort;
			break;

		default:
			break;

		}

		mFile.Seek(origin + size);
	}

	return true;

Abort:
	return false;
}


inline bool MatroskaTrack::ParseAudio(MkvFile &mFile, __int64 limit)
{
	// Parse Matroska Audio info

	int id;
	__int64 size;
	__int64 origin;

	// Clamp "limit" to prevent an endless loop
	if (limit > mFile.Size()) limit = mFile.Size();

	while (mFile.Pos() < limit)
	{
		id = mFile.ElementID();
		size = mFile.ElementSize();
		origin = mFile.Pos();

		switch (id) {

		case 0x9F:      // Channels
			nChannels = mFile.MotorolaInt((int)size);
			break;

		case 0xB5:      // SamplingFrequency
			dwSamplingFrequency = (DWORD)mFile.MotorolaFloat((int)size);
			break;

		case 0xBF:      // CRC-32
		case 0xEC:      // Void
			break;

		case 0x6264:    // BitDepth
			wBitsPerSample = (WORD)mFile.MotorolaInt((int)size);
			break;

		case 0x78B5:    // OutputSamplingFrequency
			dwOutputSamplingFrequency = (DWORD)mFile.MotorolaFloat((int)size);
			break;

		case 0x7D7B:    // ChannelPositions
		default:
			break;

		}

		mFile.Seek(origin + size);
	}

	return true;
}


inline bool MatroskaTrack::ParseVideo(MkvFile &mFile, __int64 limit)
{
	int id;
	__int64 size;
	__int64 origin;

	// Clamp "limit" to prevent an endless loop
	if (limit > mFile.Size()) limit = mFile.Size();

	while (mFile.Pos() < limit)
	{
		id = mFile.ElementID();
		size = mFile.ElementSize();
		origin = mFile.Pos();

		switch (id) {

		case 0x9A:      // FlagInterlaced
			break;

		case 0xB0:      // PixelWidth
			iPixelWidth = mFile.MotorolaInt((int)size);
			if (iPixelWidth <= 0 || iPixelWidth > MKV_MAX_WIDTH) goto Abort;
			break;

		case 0xBA:      // PixelHeight
			iPixelHeight = mFile.MotorolaInt((int)size);
			if (iPixelHeight <= 0 || iPixelHeight > MKV_MAX_HEIGHT) goto Abort;
			break;

		case 0x53B8:    // StereoMode
		case 0x54AA:    // PixelCropBottom
			break;

		case 0x54B0:    // DisplayWidth
		case 0x54B2:    // DisplayUnit
		case 0x54B3:    // AspectRatioType
		case 0x54BA:    // DisplayHeight
		case 0x54BB:    // PixelCropTop
		case 0x54CC:    // PixelCropLeft
		case 0x54DD:    // PixelCropRight
		case 0x2383E3:  // FrameRate
		case 0x2EB524:  // ColourSpace
		case 0x2FB523:  // GammaValue
		default:
			break;

		}

		mFile.Seek(origin + size);
	}

	return true;
	
Abort:
	return false;
}


int MatroskaTrack::GetCodecID(const char *CodecID)
{
	int ret = 0;

	if (CodecID != NULL)
	{
		int cmp;

		while (true)
		{

/*************************************************************************
							Audio Codec IDs
 *************************************************************************/

			cmp = strcmp(CodecID, "A_AAC");
			if (cmp == 0) { ret = A_AAC; break; }

			cmp = strcmp(CodecID, "A_AAC/MPEG2/LC");
			if (cmp == 0) { ret = A_AAC; break; }

			cmp = strcmp(CodecID, "A_AAC/MPEG2/LC/SBR");
			if (cmp == 0) { ret = A_AAC; break; }

			cmp = strcmp(CodecID, "A_AAC/MPEG2/MAIN");
			if (cmp == 0) { ret = A_AAC; break; }

			cmp = strcmp(CodecID, "A_AAC/MPEG2/SSR");
			if (cmp == 0) { ret = A_AAC; break; }

			cmp = strcmp(CodecID, "A_AAC/MPEG4/LC");
			if (cmp == 0) { ret = A_AAC; break; }

			cmp = strcmp(CodecID, "A_AAC/MPEG4/LC/SBR");
			if (cmp == 0) { ret = A_AAC; break; }

			cmp = strcmp(CodecID, "A_AAC/MPEG4/LTP");
			if (cmp == 0) { ret = A_AAC; break; }

			cmp = strcmp(CodecID, "A_AAC/MPEG4/MAIN");
			if (cmp == 0) { ret = A_AAC; break; }

			cmp = strcmp(CodecID, "A_AAC/MPEG4/SSR");
			if (cmp == 0) { ret = A_AAC; break; }

			cmp = strcmp(CodecID, "A_AC3");
			if (cmp == 0) { ret = A_AC3; break; }

			cmp = strcmp(CodecID, "A_AC3/BSID9");
			if (cmp == 0) { ret = A_AC3; break; }

			cmp = strcmp(CodecID, "A_AC3/BSID10");
			if (cmp == 0) { ret = A_AC3; break; }

#ifdef A_EAC3
			cmp = strcmp(CodecID, "A_EAC3");
			if (cmp == 0) { ret = A_EAC3; break; }
#endif
#ifdef A_FLAC
			cmp = strcmp(CodecID, "A_FLAC");
			if (cmp == 0) { ret = A_FLAC; break; }
#endif
			cmp = strcmp(CodecID, "A_MPEG/L2");
			if (cmp == 0) { ret = A_MPEG_L2; break; }

			cmp = strcmp(CodecID, "A_MPEG/L3");
			if (cmp == 0) { ret = A_MPEG_L3; break; }

			cmp = strcmp(CodecID, "A_MS/ACM");
			if (cmp == 0) { ret = A_MS_ACM; break; }
			
			cmp = strcmp(CodecID, "A_PCM/INT/BIG");
			if (cmp == 0) { ret = A_PCM_INT_BIG; break; }

			cmp = strcmp(CodecID, "A_PCM/INT/LIT");
			if (cmp == 0) { ret = A_PCM_INT_LIT; break; }

#ifdef A_VORBIS
			cmp = strcmp(CodecID, "A_VORBIS");
			if (cmp == 0) { ret = A_VORBIS; break; }
#endif

/*************************************************************************
							Video Codec IDs
 *************************************************************************/

			cmp = strcmp(CodecID, "V_MPEG1");
			if (cmp == 0) { ret = V_MPEG1; break; }

			cmp = strcmp(CodecID, "V_MPEG2");
			if (cmp == 0) { ret = V_MPEG2; break; }

			cmp = strcmp(CodecID, "V_MPEG4/ISO/AVC");
			if (cmp == 0) { ret = V_MPEG4_ISO_AVC; break; }

			cmp = strcmp(CodecID, "V_MS/VFW/FOURCC");
			if (cmp == 0) { ret = V_MS_VFW_FOURCC; break; }

#ifdef V_QUICKTIME
			cmp = strcmp(CodecID, "V_QUICKTIME");
			if (cmp == 0) { ret = V_QUICKTIME; break; }
#endif
			cmp = strcmp(CodecID, "V_REAL/RV10");
			if (cmp == 0) { ret = V_REAL_RV10; break; }

			cmp = strcmp(CodecID, "V_REAL/RV20");
			if (cmp == 0) { ret = V_REAL_RV20; break; }

			cmp = strcmp(CodecID, "V_REAL/RV30");
			if (cmp == 0) { ret = V_REAL_RV30; break; }

			cmp = strcmp(CodecID, "V_REAL/RV40");
			if (cmp == 0) { ret = V_REAL_RV40; break; }

#ifdef V_THEORA
			cmp = strcmp(CodecID, "V_THEORA");
			if (cmp == 0) { ret = V_THEORA; break; }
#endif
			break;
		}
	}
	
	return ret;
}


bool MatroskaTrack::ParseTrackEntry(MkvFile &mFile, __int64 limit)
{
	int id;
	__int64 size;
	__int64 origin;

	// Clamp "limit" to prevent an endless loop
	if (limit > mFile.Size()) limit = mFile.Size();

	while (mFile.Pos() < limit)
	{
		id = mFile.ElementID();
		size = mFile.ElementSize();
		origin = mFile.Pos();

		switch (id) {

		// Most of these entries must only occur once.

		case 0x83:      // TrackType
			if (iTrackType) goto Abort;
			iTrackType = mFile.MotorolaInt((int)size);
			break;

		case 0x86:      // CodecID
		{
			char *CodecID;
			if (iCodecID) goto Abort;

			CodecID = mFile.MatroskaStringA((int)size);

			if (CodecID != NULL)
			{
				iCodecID = GetCodecID(CodecID);
				delete[] CodecID;

				if (iCodecID == 0)
				{
					if (iTrackType == MKV_TYPE_AUDIO) goto Abort;
					if (iTrackType == MKV_TYPE_VIDEO) goto Abort;
				}
			}

			break;
		}

		case 0x88:      // FlagDefault
		case 0x9C:      // FlagLacing
		case 0xAA:      // CodecDecodeAll
		case 0xB9:      // FlagEnabled
		case 0xBF:      // CRC-32
			break;

		case 0xD7:      // TrackNumber
			if (iTrackNumber) goto Abort;
			iTrackNumber = mFile.MotorolaInt((int)size);
			break;

		case 0xE0:      // Video
			if (iTrackType != 0 && iTrackType != MKV_TYPE_VIDEO) goto Abort;
			if (!ParseVideo(mFile, origin + size)) goto Abort;
			break;

		case 0xE1:      // Audio
			if (iTrackType != 0 && iTrackType != MKV_TYPE_AUDIO) goto Abort;
			if (!ParseAudio(mFile, origin + size)) goto Abort;
			break;

		case 0xEC:      // Void
		case 0x536E:    // Name
		case 0x537F:    // TrackOffset
		case 0x55AA:    // FlagForced
		case 0x55EE:    // MaxBlockAdditionID
			break;

		case 0x63A2:    // CodecPrivate
			delete[] pCodecPrivate;
			iCodecPrivateLen = 0;

			pCodecPrivate = new unsigned char[(size_t)size];
			if (pCodecPrivate == NULL) goto Abort;
			iCodecPrivateLen = (int)size;

			mFile.Read(pCodecPrivate, iCodecPrivateLen);
			break;

		case 0x6624:    // TrackTranslate
			break;

		case 0x6D80:    // ContentEncodings
			if (!ParseContentEncodings(mFile, origin + size)) goto Abort;
			break;

		case 0x6DE7:    // MinCache
		case 0x6DF8:    // MaxCache
		case 0x6FAB:    // TrackOverlay
		case 0x73C5:    // TrackUID
		case 0x7446:    // AttachmentLink
			break;

		case 0x22B59C:  // Language
		{
			char *lang = mFile.MatroskaStringA((int)size);
			if (lang != NULL)
			{
				iLanguageID = InputFileMKV::lang_to_id(lang);
				delete[] lang;
			}
			break;
		}

		case 0x23314F:  // TrackTimecodeScale
		case 0x23E383:  // DefaultDuration
		case 0x258688:  // CodecName
		case 0x26B240:  // CodecDownloadURL
		case 0x3A9697:  // CodecSettings
		case 0x3B4040:  // CodecInfoURL
		default:
			break;

		}

		mFile.Seek(origin + size);
	}

	return true;

Abort:
	return false;
}


///////////////////////////////////////////////////////////////////////
//			Create WAVEFORMATs for supported audio formats
///////////////////////////////////////////////////////////////////////


bool MatroskaTrack::ReadStream(MkvFile &mFile, void *lpBuffer, int bytes)
{
	// Read bytes from the beginning of the stream
	
	__int64 prior;
	DWORD   i;
	bool    ret;
	int     tc;
	
	prior = mFile.Pos();
	ret   = false;

	if (bytes == 0)
	{
		ret = true;
		goto Abort;
	}
	if (bytes < 0) goto Abort;
	if (lpBuffer == NULL) goto Abort;

	for (i = 0; i < dwActualLength; ++i)
	{
		if (iCompSettingsLen > 0)
		{
			tc = iCompSettingsLen;

			if (tc > bytes) tc = bytes;
			memcpy(lpBuffer, pCompSettings, tc);

			bytes -= tc;
			if (bytes <= 0)
			{
				ret = true;
				break;
			}
			lpBuffer = (char *)lpBuffer + tc;
		}

		mFile.Seek(sample_list[i].pos);
		tc = sample_list[i].size & (~MKV_KEYFRAME);

		if (tc > bytes) tc = bytes;
		if (mFile.Read(lpBuffer, tc) != tc) break;
		
		bytes -= tc;
		if (bytes <= 0)
		{
			ret = true;
			break;
		}
		lpBuffer = (char *)lpBuffer + tc;
	}

Abort:
	mFile.Seek(prior);
	return ret;
}


inline bool MatroskaTrack::FormatPCM()
{
	WAVEFORMATEX *wfex;

	if (dwSamplingFrequency == 0) goto Abort;
	if (iCompSettingsLen != 0) goto Abort;
	if (wBitsPerSample != 8 && wBitsPerSample != 16) goto Abort;

	// Channels element not provided?
	if (nChannels == 0) nChannels = 1;
	
	delete[] pFormat;
	iFormatLen = 0;

	pFormat = new unsigned char[sizeof(WAVEFORMATEX)];
	if (pFormat == NULL) goto Abort;
	iFormatLen = sizeof(WAVEFORMATEX);

	wfex = (WAVEFORMATEX *)pFormat;

	wfex->wFormatTag      = WAVE_FORMAT_PCM;
	wfex->nChannels       = (WORD)nChannels;
	wfex->nSamplesPerSec  = dwSamplingFrequency;
	wfex->wBitsPerSample  = wBitsPerSample;
	wfex->nBlockAlign     = wfex->nChannels * (wfex->wBitsPerSample / 8);
	wfex->nAvgBytesPerSec = wfex->nSamplesPerSec * wfex->nBlockAlign;
	wfex->cbSize = 0;

	return true;
	
Abort:
	return false;
}


inline bool MatroskaTrack::FormatWMA()
{
	// Microsoft Windows Media Audio v1 and v2

	static const char WmaPatch[] = "1A0F78F0-EC8A-11d2-BBBE-006008320064";

	WAVEFORMATEX *wfex = (WAVEFORMATEX *)pCodecPrivate;

	switch (wfex->wFormatTag) {

	case WAVE_FORMAT_MSAUDIO1:      // Windows Media Audio v1
		if (iCodecPrivateLen != (sizeof(WAVEFORMATEX) + 4)) goto Abort;
		break;

	case WAVE_FORMAT_MSAUDIO2:      // Windows Media Audio v2
		if (iCodecPrivateLen != (sizeof(WAVEFORMATEX) + 10)) goto Abort;
		break;

	case WAVE_FORMAT_MSAUDIO_PRO:   // Windows Media Audio Pro
		if (iCodecPrivateLen != (sizeof(WAVEFORMATEX) + 18)) goto Abort;
		break;
	}

	pFormat = new unsigned char[iCodecPrivateLen + sizeof(WmaPatch)];
	if (pFormat == NULL) goto Abort;

	iFormatLen = iCodecPrivateLen + sizeof(WmaPatch);

	memcpy(pFormat, pCodecPrivate, iCodecPrivateLen);
	memcpy(pFormat + iCodecPrivateLen, WmaPatch, sizeof(WmaPatch));

	wfex = (WAVEFORMATEX *)pFormat;
	wfex->cbSize += sizeof(WmaPatch);

	return true;

Abort:
	return false;
}


bool MatroskaTrack::FormatMPEG(MkvFile &mFile)
{
	// MPEG audio Layers II and III

	DWORD dwAvgBytesPerSec;

	// (a / b) / c = a / (b * c)
	// a / (b / c) = (a * c) / b

	dwAvgBytesPerSec = (DWORD)((double)ui64TotalBytes / Duration + 0.5);

	// Derive these if necessary from the MPEG audio header

	if (nChannels == 0 || dwSamplingFrequency == 0)
	{

// ........ ........ 1111.... 11111111   sync (all bits set)
// ........ ........ ....!... ........   version (1=MPEG-1, 0=MPEG-2)
// ........ ........ .....!!. ........   layer (00=reserved,11=I,10=II,01=III)
// ........ ........ .......x ........   protected by CRC (0=yes, 1=no)
// ........ xxxx.... ........ ........   bitrate_index (can't be 0000 or 1111)
// ........ ....!!.. ........ ........   sampling_freq (can't be 11)
// ........ ......x. ........ ........   padding_bit (0=no, 1=yes)
// ........ .......x ........ ........   private_bit (not used)
// xx...... ........ ........ ........   (00=stereo,01=joint,10=dual,11=mono)
// ..xx.... ........ ........ ........   mode_ext (only if joint stereo)
// ....x... ........ ........ ........   copyright_bit (0=no, 1=yes)
// .....x.. ........ ........ ........   original_or_copy (0=original, 1=copy)
// ......xx ........ ........ ........   emphasis (00=none,10=reserved,01=50/15ms,11=CCIT J.17)

		unsigned long hdr;
		
		if (!ReadStream(mFile, &hdr, sizeof(hdr))) goto Abort;

		// Some sanity checks
		if ((hdr & 0x0000F0FF) != 0x0000F0FF) goto Abort;
		if ((hdr & 0x00000600) == 0x00000000) goto Abort;
		if ((hdr & 0x00F00000) == 0x00000000) goto Abort;
		if ((hdr & 0x00F00000) == 0x00F00000) goto Abort;
		if ((hdr & 0x000C0000) == 0x000C0000) goto Abort;

		if (dwSamplingFrequency == 0)
		{
			switch (hdr & 0x000C0000) {
				case 0x00000000: dwSamplingFrequency = 44100; break;
				case 0x00040000: dwSamplingFrequency = 48000; break;
				case 0x00080000: dwSamplingFrequency = 32000; break;
			}
			if ((hdr & 0x00000800) == 0) dwSamplingFrequency >>= 1;
		}

		if (nChannels == 0)
		{
			nChannels = ((hdr & 0xC0000000) == 0xC0000000)? 1: 2;
		}
	}

	switch (iCodecID) {

	case A_MPEG_L2:
	{
		MPEG1WAVEFORMAT *wf;

		pFormat = new unsigned char[sizeof(MPEG1WAVEFORMAT)];
		if (pFormat == NULL) goto Abort;

		iFormatLen = sizeof(MPEG1WAVEFORMAT);

		wf = (MPEG1WAVEFORMAT *)pFormat;

		wf->wfx.wFormatTag      = WAVE_FORMAT_MPEG;
		wf->wfx.nChannels       = nChannels;
		wf->wfx.nSamplesPerSec  = dwSamplingFrequency;
		wf->wfx.nAvgBytesPerSec = dwAvgBytesPerSec;
		wf->wfx.nBlockAlign     = (WORD)dwMaxSample;
		wf->wfx.wBitsPerSample  = 0;
		wf->wfx.cbSize          = 22;

		// This is mostly guesswork
		wf->fwHeadLayer         = 2;
		wf->dwHeadBitrate       = (DWORD)
			((double)ui64TotalBytes / (Duration * 125.0) + 0.5);
		
		if (wf->wfx.nChannels > 1)
			wf->fwHeadMode      = ACM_MPEG_STEREO | ACM_MPEG_JOINTSTEREO;
		else
			wf->fwHeadMode      = ACM_MPEG_SINGLECHANNEL;

		wf->fwHeadModeExt       = 0x000F;
		wf->wHeadEmphasis       = TRUE;
		wf->fwHeadFlags         = ACM_MPEG_ID_MPEG1 | ACM_MPEG_PROTECTIONBIT;
		wf->dwPTSLow            = 0;
		wf->dwPTSHigh           = 0;

		break;
	}

	case A_MPEG_L3:
	{
		MPEGLAYER3WAVEFORMAT *wf;

		pFormat = new unsigned char[sizeof(MPEG1WAVEFORMAT)];
		if (pFormat == NULL) goto Abort;

		iFormatLen = sizeof(MPEGLAYER3WAVEFORMAT);

		wf = (MPEGLAYER3WAVEFORMAT *)pFormat;

		wf->wfx.wFormatTag      = WAVE_FORMAT_MPEGLAYER3;
		wf->wfx.nChannels       = nChannels;
		wf->wfx.nSamplesPerSec  = dwSamplingFrequency;
		wf->wfx.nAvgBytesPerSec = dwAvgBytesPerSec;
		wf->wfx.nBlockAlign     = (WORD)dwMaxSample;
		wf->wfx.wBitsPerSample  = 0;
		wf->wfx.cbSize          = MPEGLAYER3_WFX_EXTRA_BYTES;

		wf->wID                 = MPEGLAYER3_ID_MPEG;
		wf->fdwFlags            = MPEGLAYER3_FLAG_PADDING_ISO;
		wf->nBlockSize          = (WORD)dwMaxSample;
		wf->nFramesPerBlock     = 1;
		wf->nCodecDelay         = 0;

		break;
	}
	
	}

	return true;

Abort:
	return false;
}


inline bool MatroskaTrack::CreateASC(int sr, int ch)
{
	// Create AAC AudioSpecificConfig
	
	unsigned char *dst;

	if (ch < 1 || (ch > 6 && ch != 8)) goto Abort;
	if (sr <= 0 || sr > 0xFFFFFF) goto Abort;

	pCodecPrivate = new unsigned char[5];
	if (pCodecPrivate == NULL) goto Abort;
	iCodecPrivateLen = 2;
	dst = pCodecPrivate;

	// AudioSpecificConfig, 2 bytes
	
	//   LSB      MSB
	// xxxxx... ........  object type (00010 = AAC LC)
	// .....xxx x.......  sampling frequency index (0 to 11)
	// ........ .xxxx...  channels configuration (1 to 7)
	// ........ .....x..  frame length flag (0 = 1024, 1 = 960)
	// ........ ......x.  depends on core coder (0)
	// ........ .......x  extensions flag (0)

	dst[0] = 0x10;
	dst[1] = 0x00;
		
	// .....000 0....... = 96000
	// .....000 1....... = 88200
	// .....001 0....... = 64000
	// .....001 1....... = 48000
	// .....010 0....... = 44100
	// .....010 1....... = 32000
	// .....011 0....... = 24000
	// .....011 1....... = 22050
	// .....100 0....... = 16000
	// .....100 1....... = 12000
	// .....101 0....... = 11025
	// .....101 1....... =  8000

	switch (sr) {
	case 96000: break;
	case 88200: dst[1] |= 0x80; break;
	case 64000: dst[0] |= 0x01; break;
	case 48000: dst[0] |= 0x01;
	            dst[1] |= 0x80; break;
	case 44100: dst[0] |= 0x02; break;
	case 32000: dst[0] |= 0x02;
	            dst[1] |= 0x80; break;
	case 24000: dst[0] |= 0x03; break;
	case 22050: dst[0] |= 0x03;
	            dst[1] |= 0x80; break;
	case 16000: dst[0] |= 0x04; break;
	case 12000: dst[0] |= 0x04;
	            dst[1] |= 0x80; break;
	case 11025: dst[0] |= 0x05; break;
	case  8000: dst[0] |= 0x05;
	            dst[1] |= 0x80; break;

	default:

	// 00010... ........ ........ ........ ........ object type
	// .....111 1....... ........ ........ ........ (1111 = custom)
	// ........ .xxxxxxx xxxxxxxx xxxxxxxx x....... sampling rate
	// ........ ........ ........ ........ .xxxx... channel config

		dst[0] |= 0x07;
		dst[1]  = 0x80 | ((sr >> 17) & 0x7F);
		dst[2]  = (sr >> 9) & 0xFF;
		dst[3]  = (sr >> 1) & 0xFF;
		dst[4]  = (sr & 1) << 7;
		dst += 3;

		iCodecPrivateLen += 3;
		break;
	}
	
	++dst;

	// ........ .0001... = 1
	// ........ .0010... = 2
	// ........ .0011... = 3
	// ........ .0100... = 4
	// ........ .0101... = 5
	// ........ .0110... = 6
	// ........ .0111... = 8

	switch (ch) {
	case 1: *dst |= 0x08; break;
	case 2: *dst |= 0x10; break;
	case 3: *dst |= 0x18; break;
	case 4: *dst |= 0x20; break;
	case 5: *dst |= 0x28; break;
	case 6: *dst |= 0x30; break;
	case 8: *dst |= 0x38; break;
	}

	return true;

Abort:
	return false;
}


inline bool MatroskaTrack::FormatAAC(MkvFile &mFile)
{
	// Advanced Audio Codec

	static const unsigned char chan_map[16] = {
		0, 1, 2, 3, 4, 5, 6, 8, 0, 0, 0, 0, 0, 0, 0, 0
	};

	static const unsigned long freq_map[16] = {
		96000, 88200, 64000, 48000, 44100, 32000, 24000, 22050,
		16000, 12000, 11025,  8000,  7350,     0,     0,     0
	};

// Example of LC SBR ADTS header (22050 -> 44100), 6 channels
// sf_index = 7 (22050), channel configuration = 6
// FF F1 5D 80 0B BF FC

			// 00: 11111111   syncword
			// 01: 1111....   syncword (continued)
			//     ....0...   id (1 = MPEG-2, 0 = MPEG-4)
			//     .....00.   layer
			//     .......1   protection absent
			// 02: 01......   profile
			//     ..0111..   sf_index
			//     ......0.   private bit
			//     .......1   channel configuration
			// 03: 10......   channel configuration (continued)
			//     ..0.....   original
			//     ...0....   home
			//     ....0...   copyright identification bit
			//     .....0..   copyright identification start
			//     ......00   frame length
			// 04: 00001011   frame length (continued)
			// 05: 101.....   frame length (continued)
			//     ...11111   buffer fullness
			// 06: 111111..   buffer fullness (continued)
			//     ......00   raw data blocks in frame

	WAVEFORMATEX *wfex;

	if (pCodecPrivate == NULL)
	{
		if (dwSamplingFrequency == 0 || nChannels == 0)
		{
			// If there is no ADTS header we're screwed!
			unsigned long hdr;

			if (!ReadStream(mFile, &hdr, sizeof(hdr))) goto Abort;

// ........ ........ 1111.... 11111111  syncword
// ........ ........ ....0... ........  id (1 = MPEG-2, 0 = MPEG-4)
// ........ ........ .....00. ........  layer
// ........ ........ .......x ........  protection absent
// ........ xx...... ........ ........  profile
// ........ ..xxxx.. ........ ........  sf_index
// ........ ......x. ........ ........  private bit
// xx...... .......x ........ ........  channel configuration

			if ((hdr & 0x0000FEFF) != 0x0000F0FF) goto Abort;

			if (dwSamplingFrequency == 0)
			{
				dwSamplingFrequency = freq_map[(hdr >> 18) & 15];
				if (dwSamplingFrequency == 0) goto Abort;
			}
			
			if (nChannels == 0)
			{
				nChannels = chan_map[((hdr >> 14) & 4) + (hdr >> 30)];
				if (nChannels == 0) goto Abort;
			}
		}

		if (!CreateASC(dwSamplingFrequency, nChannels)) goto Abort;

		// Implicit SBR signalling?
		if (dwSamplingFrequency <= 24000 && dwOutputSamplingFrequency == 0)
		{
			dwOutputSamplingFrequency = dwSamplingFrequency * 2;
		}
	}
	else
	{
		// Derive from AudioSpecificConfig

		//   LSB      MSB
		// xxxxx... ........  object type (00010 = AAC LC)
		// .....xxx x.......  sampling frequency index (0 to 11)
		// ........ .xxxx...  channels configuration (1 to 7)
		// ........ .....x..  frame length flag (0 = 1024, 1 = 960)
		// ........ ......x.  depends on core coder (0)
		// ........ .......x  extensions flag (0)

		if (dwSamplingFrequency == 0)
		{
			dwSamplingFrequency = (pCodecPrivate[0] << 8) + pCodecPrivate[1];
			dwSamplingFrequency = freq_map[(dwSamplingFrequency >> 7) & 15];
			if (dwSamplingFrequency == 0) goto Abort;
		}

		if (nChannels == 0)
		{
			nChannels = chan_map[(pCodecPrivate[1] >> 3) & 15];
			if (nChannels == 0) goto Abort;

			if (nChannels == 1)
			{
				// Try to detect HE AAC v2 with PS (what a mess)...

				// 00: 00010...  object type (00010 = LC)
				//     .....xxx  sampling frequency index (0 to 11)
				// 01: x.......  sampling frequency index continued
				//     .xxxx...  channels configuration (1 to 7)
				//     .....x..  frame length flag (0 = 1024, 1 = 960)
				//     ......0.  depends on core coder (0)
				//     .......0  extensions flag (0)
				// 02: 01010110  syncExtensionType (0x02B7)
				// 03: 111.....  syncExtensionType continued
				//     ...00101  extensionAudioObjectType (5)
				// 04: 1.......  sbrPresentFlag (0 = no, 1 = yes)
				//     .xxxx...  extensionSamplingFrequencyIndex
				//     .....101  syncExtensionType (0x0548)
				// 05: 01001000  syncExtensionType continued
				// 06: x.......  psPresentFlag (0 = no, 1 = yes)

				if (
				      (   iCodecPrivateLen >= 7
				      &&  pCodecPrivate[2] == 0x56
				      &&  pCodecPrivate[3] == 0xE5
				      && (pCodecPrivate[4] & 0x87) == 0x85
				      &&  pCodecPrivate[5] == 0x48
				      && (pCodecPrivate[6] & 0x80) == 0x80
				      )
				   ||
				      (   dwOutputSamplingFrequency != 0
				      &&  dwOutputSamplingFrequency != dwSamplingFrequency
				      )
				   )
				{
					++nChannels;
				}
			}
		}

		// Implicit SBR signalling?
		if (iCodecPrivateLen < 4 && dwSamplingFrequency <= 24000 && dwOutputSamplingFrequency == 0)
		{
			dwOutputSamplingFrequency = dwSamplingFrequency * 2;
		}
	}

	if (dwOutputSamplingFrequency == 0)
	{
		dwOutputSamplingFrequency = dwSamplingFrequency;
	}

	pFormat = new unsigned char[sizeof(WAVEFORMATEX) + iCodecPrivateLen];
	if (pFormat == NULL) goto Abort;
	
	iFormatLen = sizeof(WAVEFORMATEX) + iCodecPrivateLen;

	wfex = (WAVEFORMATEX *)pFormat;

	wfex->wFormatTag      = WAVE_FORMAT_RAW_AAC1;
	wfex->nChannels       = nChannels;
	wfex->nSamplesPerSec  = dwOutputSamplingFrequency;
	wfex->nBlockAlign     = (WORD)dwMaxSample;
	wfex->nAvgBytesPerSec = (DWORD)((double)ui64TotalBytes / Duration + 0.5);
	wfex->wBitsPerSample  = 0;
	wfex->cbSize          = iCodecPrivateLen;

	memcpy(pFormat + sizeof(WAVEFORMATEX), pCodecPrivate, iCodecPrivateLen);

	return true;

Abort:
	return false;
}


int MatroskaTrack::ac3_sync_info(unsigned long hdr, DWORD *srate, int *brate, int *chans)
{
	static const unsigned short rate[19] = {
		 32,  40,  48,  56,  64,  80,  96, 112, 128,
		160, 192, 224, 256, 320, 384, 448, 512, 576, 640
	};

	static const unsigned short lfeon[8] = {
		0x1000, 0x1000, 0x1000, 0x0400,	0x0400, 0x0100, 0x0400, 0x0100
	};

	static const unsigned char chan[8] = { 2, 1, 2, 3, 3, 4, 4, 5 };

	static const unsigned char halfrate[12] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 2, 3 };

	if (hdr < 0xC0000000)
	{
		const int frmsizecod = (hdr >> 24) & 63;

		if (frmsizecod < 38)
		{
			const int bsid = (hdr >> 19) & 31;

			if (bsid < 12)
			{
				const int acmod   = (hdr >> 13) & 7;
				const int half    = halfrate[bsid];
				const int bitrate = rate[frmsizecod >> 1];

				if (chans) *chans = chan[acmod] + !!(hdr & lfeon[acmod]);
				if (brate) *brate = bitrate >> half;

				switch (hdr >> 30)
				{
					case 0:
						if (srate) *srate = 48000 >> half;
						return (4 * bitrate);

					case 1:
						if (srate) *srate = 44100 >> half;
						return (2 * (320 * bitrate / 147 + (frmsizecod & 1)));

					case 2:
						if (srate) *srate = 32000 >> half;
						return (6 * bitrate);
				}
			}
		}
	}

	return 0;
}


inline bool MatroskaTrack::FormatAC3(MkvFile &mFile)
{
	// Dolby Digital

	WAVEFORMATEX *wfex;

	// Derive these if necessary from the AC-3 header
	if (dwSamplingFrequency == 0 || nChannels == 0)
	{
		unsigned char buf[8];
		unsigned long hdr;
		int framesize;

		if (!ReadStream(mFile, buf, 8)) goto Abort;

		if (buf[0] != 0x0B)
		{
			// Intel AC-3 ?
			if (buf[0] != 0x77) goto Abort;
			if (buf[1] != 0x0B) goto Abort;

			hdr = buf[5];
			hdr = (hdr << 8) + buf[4];
			hdr = (hdr << 8) + buf[7];
			hdr = (hdr << 8) + buf[6];
		}
		else
		{
			// Motorola AC-3
			if (buf[1] != 0x77) goto Abort;

			hdr = buf[4];
			hdr = (hdr << 8) + buf[5];
			hdr = (hdr << 8) + buf[6];
			hdr = (hdr << 8) + buf[7];
		}

		framesize = ac3_sync_info(hdr, &dwSamplingFrequency, NULL, &nChannels);
		if (framesize <= 0) goto Abort;
	}

	switch (dwSamplingFrequency) {
		case  8000: case 11025: case 12000:
		case 16000: case 22050: case 24000:
		case 32000: case 44100: case 48000:
			break;
		default:
			goto Abort;
	}
	
	if (nChannels < 1 || nChannels > 6) goto Abort;

	pFormat = new unsigned char[sizeof(WAVEFORMATEX)];
	if (pFormat == NULL) goto Abort;
	
	iFormatLen = sizeof(WAVEFORMATEX);

	wfex = (WAVEFORMATEX *)pFormat;

	wfex->wFormatTag      = WAVE_FORMAT_AC3;
	wfex->nChannels       = nChannels;
	wfex->nSamplesPerSec  = dwSamplingFrequency;
	wfex->nBlockAlign     = (WORD)dwMaxSample;
	wfex->nAvgBytesPerSec = (DWORD)((double)ui64TotalBytes / Duration + 0.5);
	wfex->wBitsPerSample  = 0;
	wfex->cbSize          = 0;

	return true;

Abort:
	return false;
}


#ifdef A_VORBIS

inline bool MatroskaTrack::FormatVorbis()
{
	// Vorbis I

	const unsigned char *hdr;
	WAVEFORMATEX *wfex;
	int remains;
	int sizes;

	if (iCodecPrivateLen < 33) goto Abort;

	// Get the channels and sample rate from the identification header.
	// The vorbis headers are Xiph-lace encoded (aargh...)!

	// Britney Spears -Circus.mkv has:
	// 02 1e 2d 01 76 6f 72 62 69 73 00 00 00 00 02 44  ..-.vorbis.....D
	// ac 00 00 00 00 00 00 00 71 02 00 00 00 00 00 b8  ........q.......
	// 01 03 76 6f 72 62 69 73 1d 00 00 00 58 69 70 68  ..vorbis....Xiph
	// 2e 4f 72 67 20 6c 69 62 56 6f 72 62 69 73 20 49  .Org libVorbis I
	// 20 32 30 30 37 30 36 32 32 00 00 00 00 01 05 76   20070622......v
	// 6f 72 62 69 73 29 42 43 56 01 00 08 00 00 00 31  orbis)BCV......1

	// 02 (number of sizes)
	// 1e (size of first block)
	// 2d (size of second block)
	// first block (identification header, 30 bytes):
	// 01 76 6f 72 62 69 73 00 00 00 00 02 44 ac 00 00  .vorbis.....D...
	// 00 00 00 00 00 71 02 00 00 00 00 00 b8 01        .....q........
	// second block (comment header, varies):
	// 03 76 6f 72 62 69 73 1d 00 00 00 58 69 70 68 2e  .vorbis....Xiph.
	// 4f 72 67 20 6c 69 62 56 6f 72 62 69 73 20 49 20  Org libVorbis I 
	// 32 30 30 37 30 36 32 32 00 00 00 00 01           20070622.....
	// third block (setup header, varies):
	// 05 76 6f 72 62 69 73 29 42 43 56 01 00 08 00 00  .vorbis)BCV.....
	// ...
	// Note: the order of these headers is mandatory.

	// Skip past the Xiph size fields
	sizes   = pCodecPrivate[0];
	hdr     = pCodecPrivate + 1;
	remains = iCodecPrivateLen - 1;

	while (sizes-- > 0)
	{
		unsigned char sz = *hdr++;
		if (--remains < 30) goto Abort;

		while (sz == 255)
		{
			sz = *hdr++;
			if (--remains < 30) goto Abort;
		}
	}

	// Now we must be at the Vorbis identification header.

	// 00: 01                 packet type (identification header)
	// 01: 76 6F 72 62 69 73  string 'vorbis'
	// 07: 00 00 00 00        vorbis_version (0)
	// 11: xx                 audio_channels (not 0)
	// 12: xx xx xx xx        audio_sample_rate (not 0)

	if (memcmp(hdr, "\001vorbis\0\0\0\0", 11) != 0) goto Abort;

	nChannels = hdr[11];
	dwSamplingFrequency = *(DWORD *)(hdr + 12);

	if (nChannels == 0 || dwSamplingFrequency == 0) goto Abort;

	pFormat = new unsigned char[sizeof(WAVEFORMATEX) + iCodecPrivateLen];
	if (pFormat == NULL) goto Abort;

	iFormatLen = sizeof(WAVEFORMATEX) + iCodecPrivateLen;

	wfex = (WAVEFORMATEX *)pFormat;

	wfex->wFormatTag      = WAVE_FORMAT_VORBIS_I;
	wfex->nChannels       = nChannels;
	wfex->nSamplesPerSec  = dwSamplingFrequency;
	wfex->nBlockAlign     = (WORD)dwMaxSample;
	wfex->nAvgBytesPerSec = (DWORD)((double)ui64TotalBytes / Duration + 0.5);
	wfex->wBitsPerSample  = 0;
	wfex->cbSize          = (WORD)iCodecPrivateLen;

	memcpy(pFormat + sizeof(WAVEFORMATEX), pCodecPrivate, iCodecPrivateLen);

	return true;

Abort:
	return false;
}

#endif	// A_VORBIS


#ifdef A_FLAC

inline bool MatroskaTrack::FormatFLAC()
{
	// Free Lossless Audio Codec

	WAVEFORMATEX *wfex;

	if (iCodecPrivateLen < 1) goto Abort;

	pFormat = new unsigned char[sizeof(WAVEFORMATEX) + iCodecPrivateLen];
	if (pFormat == NULL) goto Abort;
	iFormatLen = sizeof(WAVEFORMATEX) + iCodecPrivateLen;

	wfex = (WAVEFORMATEX *)pFormat;

	wfex->wFormatTag      = WAVE_FORMAT_FLAC;
	wfex->nChannels       = nChannels;
	wfex->nSamplesPerSec  = dwSamplingFrequency;
	wfex->nBlockAlign     = (WORD)dwMaxSample;
	wfex->nAvgBytesPerSec = (DWORD)((double)ui64TotalBytes / Duration + 0.5);
	wfex->wBitsPerSample  = 0;
	wfex->cbSize          = iCodecPrivateLen;

	memcpy(pFormat + sizeof(WAVEFORMATEX), pCodecPrivate, iCodecPrivateLen);

	return true;

Abort:
	return false;
}

#endif	// A_FLAC


bool MatroskaTrack::FinishAudio(MkvFile &mFile)
{
	// Initialize audio track properties.
	// Beware: nChannels is sometimes not present!

	switch (iCodecID) {

	case A_AAC:
		if (!FormatAAC(mFile)) goto Abort;
		break;

	case A_AC3:
		if (!FormatAC3(mFile)) goto Abort;
		break;

#ifdef A_FLAC
	case A_FLAC:
		if (!FormatFLAC()) goto Abort;
		break;
#endif

	case A_MPEG_L2:
	case A_MPEG_L3:
		if (!FormatMPEG(mFile)) goto Abort;
		break;

	case A_MS_ACM:
	{
		WAVEFORMAT *wf;

		if (pCodecPrivate == NULL) goto Abort;
		if (iCodecPrivateLen < sizeof(WAVEFORMAT)) goto Abort;

		wf = (WAVEFORMAT *)pCodecPrivate;

		if (nChannels == 0) nChannels = wf->nChannels;		
		if (dwSamplingFrequency == 0) dwSamplingFrequency = wf->nSamplesPerSec;

		switch (wf->wFormatTag) {

		case WAVE_FORMAT_MSAUDIO1:      // Windows Media Audio v1
		case WAVE_FORMAT_MSAUDIO2:      // Windows Media Audio v2
		case WAVE_FORMAT_MSAUDIO_PRO:   // Windows Media Audio Pro
			if (!FormatWMA()) goto Abort;
			break;

		// If it's MPEG, generate our own WAVE format.

		case WAVE_FORMAT_MPEG:          // MPEG Layers I, II
			iCodecID = A_MPEG_L2;
			if (!FormatMPEG(mFile)) goto Abort;
			break;

		case WAVE_FORMAT_MPEGLAYER3:    // MPEG Layer III
			iCodecID = A_MPEG_L3;
			if (!FormatMPEG(mFile)) goto Abort;
			break;

		default:
			// Unknown format; sanitize it a bit.
			wf->nAvgBytesPerSec = (DWORD)((double)ui64TotalBytes / Duration + 0.5);
			wf->nBlockAlign     = (WORD)dwMaxSample;

			pFormat             = pCodecPrivate;
			iFormatLen          = iCodecPrivateLen;
			pCodecPrivate       = NULL;
			iCodecPrivateLen    = 0;
			break;
		}

		break;
	}

	case A_PCM_INT_LIT:
	case A_PCM_INT_BIG:
		if (!FormatPCM()) goto Abort;
		break;

#ifdef A_VORBIS
	case A_VORBIS:
		if (!FormatVorbis()) goto Abort;
		break;
#endif

	default:
		goto Abort;
	}

	// Do not treat PCM as VBR
	if (InputFileMKV::IsPCM(*this)) dwFlags &= ~MKV_IS_VBR;

	return true;

Abort:
	return false;
}


int MatroskaTrack::TransformAVCConfRec(const unsigned char *src, unsigned char *dst)
{
	// Transform the AVCDecoderConfigurationRecord
	// into something that ffdshow can understand
	// by unwrapping it into raw NALUs.  Return the
	// number of bytes in the result.  We have to
	// save this data and prepend it to all key
	// frames of the AVC stream.

	// Note: this is only performed twice per video
	// track during loading, so it need not be fast.

	int total = 0;

	__try {
		int n;

		++src;  // 8 bits, configurationVersion (00000001)
		++src;  // 8 bits, AVCProfileIndication
		++src;  // 8 bits, profile_compatibility
		++src;  // 8 bits, AVCLevelIndication
		++src;  // 6 bits, reserved (111111)
		        // 2 bits, lengthSizeMinusOne

		// 3 bits, reserved (111)
		// 5 bits, numOfSequenceParameterSets
		n = *src++ & 0x1F;

		while (n-- > 0)
		{
			// 16 bits, sequenceParameterSetLength
			int size = *src++;
			size = (size << 8) + *src++;

			if (size > 0)
			{
				// sequenceParameterSetNALUnit
				if (dst != NULL)
				{
					*dst++ = 0;
					*dst++ = 0;
					*dst++ = 0;
					*dst++ = 1;

					memcpy(dst, src, size);
					dst += size;
				}
				src   += size;
				total += 4 + size;
			}
		}

		// 8 bits, numOfPictureParameterSets
		n = *src++;

		while (n-- > 0)
		{
			// 16 bits, pictureParameterSetLength
			int size = *src++;
			size = (size << 8) + *src++;

			if (size > 0)
			{
				// pictureParameterSetNALUnit
				if (dst != NULL)
				{
					*dst++ = 0;
					*dst++ = 0;
					*dst++ = 0;
					*dst++ = 1;

					memcpy(dst, src, size);
					dst += size;
				}
				src   += size;
				total += 4 + size;
			}
		}
	}
	__except(EXCEPTION_EXECUTE_HANDLER)
	{
		total = 0;
	}

	return total;
}


inline bool MatroskaTrack::FinishVideo()
{
	if (pCodecPrivate == NULL) goto Abort;

	switch (iCodecID) {

	case V_MS_VFW_FOURCC:
		delete[] pFormat;
		pFormat = pCodecPrivate;
		iFormatLen = iCodecPrivateLen;
		pCodecPrivate = NULL;
		iCodecPrivateLen = 0;
		break;

	case V_MPEG4_ISO_AVC:
		// AVCDecoderConfigurationRecord, minimum is 7 bytes
		if (iCodecPrivateLen < 7) goto Abort;

		pFormat = new unsigned char[sizeof(BITMAPINFOHEADER)];
		if (pFormat == NULL) goto Abort;
		iFormatLen = sizeof(BITMAPINFOHEADER);
		{
			unsigned char *tmp;
			int size;

			BITMAPINFOHEADER *bmih = (BITMAPINFOHEADER *)pFormat;

			bmih->biSize          = iFormatLen;
			bmih->biWidth         = iPixelWidth;
			bmih->biHeight        = iPixelHeight;
			bmih->biPlanes        = 1;
			bmih->biBitCount      = 24;
			bmih->biCompression   = '1CVA';
			bmih->biSizeImage     = ((bmih->biWidth * 3 + 3) & -4) * bmih->biHeight;
			bmih->biXPelsPerMeter = 0;
			bmih->biYPelsPerMeter = 0;
			bmih->biClrUsed       = 0;
			bmih->biClrImportant  = 0;

			// 6 bits, reserved (111111)
			// 2 bits, lengthSizeMinusOne
			avc_len = (pCodecPrivate[4] & 3) + 1;

			size = TransformAVCConfRec(pCodecPrivate, NULL);
			if (size <= 0) goto Abort;
			tmp = new unsigned char[size];
			if (tmp == NULL) goto Abort;

			iCodecPrivateLen = TransformAVCConfRec(pCodecPrivate, tmp);
			delete[] pCodecPrivate;
			pCodecPrivate = tmp;
		}
		break;

	case V_MPEG1:
	case V_MPEG2:
	case V_QUICKTIME:
	case V_REAL_RV10:
	case V_REAL_RV20:
	case V_REAL_RV30:
	case V_REAL_RV40:
	case V_THEORA:
		pFormat = new unsigned char[sizeof(BITMAPINFOHEADER) + iCodecPrivateLen];
		if (pFormat == NULL) goto Abort;
		iFormatLen = sizeof(BITMAPINFOHEADER) + iCodecPrivateLen;
		{
			BITMAPINFOHEADER *bmih = (BITMAPINFOHEADER *)pFormat;

			bmih->biSize     = iFormatLen;
			bmih->biWidth    = iPixelWidth;
			bmih->biHeight   = iPixelHeight;
			bmih->biPlanes   = 1;
			bmih->biBitCount = 24;

			switch (iCodecID)
			{
				case V_MPEG1:     bmih->biCompression = '1GPM'; break;
				case V_MPEG2:     bmih->biCompression = '2GPM'; break;
				case V_QUICKTIME: bmih->biCompression = '3QVS'; break;
				case V_REAL_RV10: bmih->biCompression = '01VR'; break;
				case V_REAL_RV20: bmih->biCompression = '02VR'; break;
				case V_REAL_RV30: bmih->biCompression = '03VR'; break;
				case V_REAL_RV40: bmih->biCompression = '04VR'; break;
				case V_THEORA:    bmih->biCompression = 'OEHT'; break;
			}

			bmih->biSizeImage     = ((bmih->biWidth * 3 + 3) & -4) * bmih->biHeight;
			bmih->biXPelsPerMeter = 0;
			bmih->biYPelsPerMeter = 0;
			bmih->biClrUsed       = 0;
			bmih->biClrImportant  = 0;

			memcpy(pFormat + sizeof(BITMAPINFOHEADER), pCodecPrivate, iCodecPrivateLen);
		}
		break;

	default:
		goto Abort;
	}

	return true;

Abort:
	return false;
}


void MatroskaTrack::FinishStamps(MKVSampleInfo *msi, const DWORD samples)
{
	// Interpolate missing time stamps
	if (msi != NULL && samples != 0)
	{
		DWORD i;

		if (msi[0].pres_time == PTS_INVALID)
		{
			// Should not happen
			msi[0].pres_time = 0;
		}

		i = 1;
		while (i < samples)
		{
			if (msi[i].pres_time == PTS_INVALID)
			{
				// Find the next valid timestamp
				DWORD j;

				for (j = i + 1; j < samples; ++j)
				{
					if (msi[j].pres_time != PTS_INVALID)
					{
						DWORD k;

						double diff = (double)
							(msi[j].pres_time -
							msi[i - 1].pres_time)
							/ (j - (i - 1));
							
						for (k = i; k < j; ++k)
						{
							msi[k].pres_time =
								msi[i - 1].pres_time +
								(DWORD)(diff * (k - (i - 1)) + 0.5);
						}
						break;
					}
				}
				
				if (j == samples)
				{
					// No valid timestamps found after i; use an average
					
					if (i > 1)
					{
						double diff = (double)
							(msi[i - 1].pres_time -
							msi[0].pres_time) / (i - 1);
							
						for (j = i; j < samples; ++j)
						{
							msi[j].pres_time = (DWORD)
								(msi[i - 1].pres_time
								+ diff * (j - (i - 1)));
						}
					}
					else
					{
						// Only one timestamp found!
						for (j = 1; j < samples; ++j)
						{
							msi[j].pres_time = msi[0].pres_time;
						}
					}
				}
				else
				{
					i = j + 1;
				}
			}
			else
			{
				++i;
			}
		}
	}
}


bool MatroskaTrack::Finish(MkvFile &mFile)
{
	DWORD i;

	if (dv == NULL) goto Abort;

	sample_list = (MKVSampleInfo *)dv->MakeArray();
	dwActualLength = dv->Length();

	delete dv;
	dv = NULL;

	if (sample_list == NULL) goto Abort;

	if (dwActualLength == 0) goto Abort;

	if (iCodecID == 0) goto Abort;

	// Calculate total bytes and maximum sample size
	dwMaxSample = sample_list[0].size & (~MKV_KEYFRAME);
	ui64TotalBytes = dwMaxSample;

	for (i = 1; i < dwActualLength; ++i)
	{
		DWORD dwSize = sample_list[i].size & (~MKV_KEYFRAME);
		if (dwSize != dwMaxSample)
		{
			dwFlags |= MKV_HAS_VSB;
			if (dwSize > dwMaxSample) dwMaxSample = dwSize;
		}
		ui64TotalBytes += dwSize;
	}

	if (ui64TotalBytes == 0 || dwMaxSample == 0) goto Abort;

	// Adjust for header stripping
	dwMaxSample    += iCompSettingsLen;
	ui64TotalBytes += (unsigned __int64)iCompSettingsLen * (unsigned __int64)dwActualLength;

	if (dwFlags & MKV_ADD_ADTS)
	{
		// Adjust for ADTS headers
		dwMaxSample    += 7;
		ui64TotalBytes += 7UI64 * (unsigned __int64)dwActualLength;
	}

	// Interpolate any missing timestamps	
	FinishStamps(sample_list, dwActualLength);

#if 0
	if (iTrackType == MKV_TYPE_AUDIO)
	{
		FILE *file = fopen("E:\\stamps.txt", "w");
		if (file != NULL)
		{
			for (i = 0; i < dwActualLength; ++i)
			{
				if (i > 0)
				{
					fprintf(file, "%lu (%lu)\n", sample_list[i].pres_time,
						sample_list[i].pres_time - sample_list[i-1].pres_time
					);
				}
				else
				{
					fprintf(file, "%lu\n", sample_list[i].pres_time);
				}
			}
			fclose(file);
		}
	}
#endif

	// Calculate track duration in seconds
	if (dwActualLength > 1)
	{
		DWORD dwStart    = 0;
		DWORD dwFirstPTS = sample_list[0].pres_time;

		DWORD dwEnd      = dwStart;
		DWORD dwLastPTS  = dwFirstPTS;

		// Locate the earliest and the latest time stamps		
		for (i = 1; i < dwActualLength; ++i)
		{
			if (sample_list[i].pres_time < dwFirstPTS)
			{
				dwFirstPTS = sample_list[i].pres_time;
				dwStart    = i;
			}
			if (sample_list[i].pres_time > dwLastPTS)
			{
				dwLastPTS  = sample_list[i].pres_time;
				dwEnd      = i;
			}
		}

		if (dwEnd != dwStart)
		{
			// (last - first) / (end - start) = milliseconds per frame
			// duration = (milliseconds per frame * numframes) / 1000

			Duration = (double)(dwLastPTS - dwFirstPTS) / (double)(dwEnd - dwStart);
			Duration = (Duration * dwActualLength) / 1000.0;
		}
	}

	// Calculate sample rate in nanoseconds
	dwRate  = 1000000000;
	dwScale = (DWORD)((Duration * dwRate) / dwActualLength + 0.5);
	
	// Do not let zero happen!
	if (dwScale == 0)
	{
		Duration = 1.0;
		dwRate   = 1;
		dwScale  = 1;
	}

	if (iTrackType == MKV_TYPE_AUDIO) return FinishAudio(mFile);
	if (iTrackType == MKV_TYPE_VIDEO) return FinishVideo();

Abort:
	return false;
}


bool MatroskaTrack::ReadFromIndex(WinFile &idx, DWORD ver)
{
	bool bIsVBR;

	// Track properties
	idx.Read(&iTrackNumber, sizeof(iTrackNumber));
	idx.Read(&iTrackType,   sizeof(iTrackType));

	// Audio/Video info
	idx.Read(&dwSamplingFrequency,       sizeof(dwSamplingFrequency));
	idx.Read(&dwOutputSamplingFrequency, sizeof(dwOutputSamplingFrequency));
	idx.Read(&nChannels,                 sizeof(nChannels));
	idx.Read(&iPixelWidth,               sizeof(iPixelWidth));
	idx.Read(&iPixelHeight,              sizeof(iPixelHeight));

	// Codec info
	idx.Read(&iCodecID,   sizeof(iCodecID));

	idx.Read(&iFormatLen, sizeof(iFormatLen));
	if (iFormatLen > 0)
	{
		pFormat = new unsigned char[iFormatLen];
		if (pFormat == NULL) goto Abort;
		idx.Read(pFormat, iFormatLen);
	}

	idx.Read(&iCodecPrivateLen, sizeof(iCodecPrivateLen));
	if (iCodecPrivateLen > 0)
	{
		pCodecPrivate = new unsigned char[iCodecPrivateLen];
		if (pCodecPrivate == NULL) goto Abort;
		idx.Read(pCodecPrivate, iCodecPrivateLen);
	}

	idx.Read(&iCompSettingsLen, sizeof(iCompSettingsLen));
	if (iCompSettingsLen > 0)
	{
		pCompSettings = new unsigned char[iCompSettingsLen];
		if (pCompSettings == NULL) goto Abort;
		idx.Read(pCompSettings, iCompSettingsLen);
	}

	idx.Read(&avc_len,        sizeof(avc_len));

	// Sample info
	idx.Read(&dwActualLength, sizeof(dwActualLength));
	if (dwActualLength > 0)
	{
		DWORD i;

		sample_list = new MKVSampleInfo[dwActualLength];
		if (sample_list == NULL) goto Abort;
		idx.Read(sample_list, dwActualLength * sizeof(MKVSampleInfo));

		dwMaxSample = sample_list[0].size & (~MKV_KEYFRAME);
		ui64TotalBytes = dwMaxSample;
		for (i = 1; i < dwActualLength; ++i)
		{
			DWORD dwSize = sample_list[i].size & (~MKV_KEYFRAME);
			if (dwSize > dwMaxSample) dwMaxSample = dwSize;
			ui64TotalBytes += dwSize;
		}
		// Adjust for header stripping
		dwMaxSample += iCompSettingsLen;
		ui64TotalBytes += (unsigned __int64)iCompSettingsLen * (unsigned __int64)dwActualLength;
	}

	if (ver < '400X')
	{
		// dwTotalBytes; deprecated
		idx.Move(4);
	}

	idx.Read(&dwRate,       sizeof(dwRate));
	idx.Read(&dwScale,      sizeof(dwScale));
	idx.Read(&Duration,     sizeof(Duration));
	idx.Read(&bIsVBR,       sizeof(bIsVBR));

	if (bIsVBR)
		dwFlags |=  MKV_HAS_VSB;
	else
		dwFlags &= ~MKV_HAS_VSB;
	
	if (ver >= '300X')
	{
		bool bDeliverRawAAC;
		idx.Read(&bDeliverRawAAC, sizeof(bDeliverRawAAC));
		if (bDeliverRawAAC)
			dwFlags &= ~MKV_ADD_ADTS;
		else
			dwFlags |=  MKV_ADD_ADTS;
	}
	if (ver >= '400X')
	{
		idx.Read(&iLanguageID,    sizeof(iLanguageID));
	}

	if (dwFlags & MKV_ADD_ADTS)
	{
		// Adjust for ADTS headers
		dwMaxSample += 7;
		ui64TotalBytes += 7UI64 * (unsigned __int64)dwActualLength;
	}

	return true;
	
Abort:
	return false;
}


bool MatroskaTrack::WriteToIndex(WinFile &idx) const
{
	bool bIsVBR         = (dwFlags & MKV_HAS_VSB )? true: false;
	bool bDeliverRawAAC = (dwFlags & MKV_ADD_ADTS)? false: true;

	// Track properties
	idx.Write(&iTrackNumber, sizeof(iTrackNumber));
	idx.Write(&iTrackType,   sizeof(iTrackType));

	// Audio/Video info
	idx.Write(&dwSamplingFrequency,       sizeof(dwSamplingFrequency));
	idx.Write(&dwOutputSamplingFrequency, sizeof(dwOutputSamplingFrequency));
	idx.Write(&nChannels,                 sizeof(nChannels));
	idx.Write(&iPixelWidth,               sizeof(iPixelWidth));
	idx.Write(&iPixelHeight,              sizeof(iPixelHeight));

	// Codec info
	idx.Write(&iCodecID,         sizeof(iCodecID));

	idx.Write(&iFormatLen,       sizeof(iFormatLen));
	if (iFormatLen > 0)
		idx.Write(pFormat,       iFormatLen);

	idx.Write(&iCodecPrivateLen, sizeof(iCodecPrivateLen));
	if (iCodecPrivateLen > 0)
		idx.Write(pCodecPrivate, iCodecPrivateLen);

	idx.Write(&iCompSettingsLen, sizeof(iCompSettingsLen));
	if (iCompSettingsLen > 0)
	    idx.Write(pCompSettings, iCompSettingsLen);

	idx.Write(&avc_len,          sizeof(avc_len));

	// Sample info
	idx.Write(&dwActualLength,   sizeof(dwActualLength));
	if (dwActualLength > 0)
	    idx.Write(sample_list,   dwActualLength * sizeof(MKVSampleInfo));

	idx.Write(&dwRate,           sizeof(dwRate));
	idx.Write(&dwScale,          sizeof(dwScale));
	idx.Write(&Duration,         sizeof(Duration));
	idx.Write(&bIsVBR,           sizeof(bIsVBR));

	idx.Write(&bDeliverRawAAC,   sizeof(bDeliverRawAAC));
	idx.Write(&iLanguageID,      sizeof(iLanguageID));

	return true;
}
