///////////////////////////////////////////////////////////////////////
// Matroska plugin for VirtualDub 1.8.8+
// Copyright (C) 2010-2012 fccHandler
// Copyright (C) 1998-2012 Avery Lee
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////

// MKV Index handlers (part of InputFileMKV class)

#if defined(_MSC_VER) && (_MSC_VER >= 1400)
	#define _CRT_SECURE_NO_WARNINGS
#endif

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <vfw.h>
#include <stdio.h>

#include "VDInputDriver.h"
#include "VDUnknown.h"

#include "MkvFile.h"
#include "DataVector.h"
#include "InputFileMKV.h"
#include "MatroskaTrack.h"
#include "Resource.h"


wchar_t *InputFileMKV::ReplaceExtension(const wchar_t *szFile, const wchar_t *szExt)
{
	wchar_t *ret = NULL;

	if (szFile != NULL && szExt != NULL)
	{
		int flen = lstrlenW(szFile);
		int elen = lstrlenW(szExt);

		if (flen > 0 && elen > 0)
		{
			int ext;
			for (ext = flen; ext > 0; )
			{
				--ext;
				if (szFile[ext] == L'.') break;
			}
		
			// ext points to the dot
			if (ext == 0) ext = flen;
				
			ret = new wchar_t[ext + elen + 2];
			if (ret != NULL)
			{
				memcpy(ret, szFile, ext * sizeof(wchar_t));
				if (szExt[0] != '.') ret[ext++] = L'.';
				memcpy(ret + ext, szExt, elen * sizeof(wchar_t));
				ret[ext + elen] = L'\0';
			}
		}
	}
	
	return ret;
}


wchar_t *InputFileMKV::GetFileNameFromIndex(const wchar_t *szIndex)
{
	WinFile   idx;
	wchar_t*  ret = NULL;

	if (idx.Open(szIndex))
	{
		DWORD ver[2];
		idx.Read(ver, sizeof(ver));
		
		if (ver[0] == 'DIKM' && (ver[1] & 0xF8FFFFFF) == '000X' && ver[1] >= '200X')
		{
			// Read file title from index file (version >= 2 only!)
			const wchar_t *title = GetFileTitle(szIndex);
			int path;
			int len;

			idx.Read(&len, sizeof(len));
			path = (int)(title - szIndex);

			ret = new wchar_t[path + len + 1];
			if (ret != NULL)
			{
				memcpy(ret, szIndex, path * sizeof(wchar_t));
				idx.Read(ret + path, len * sizeof(wchar_t));
				ret[path + len] = L'\0';
			}
		}
	}

	if (ret == NULL)
	{
		// Version 1 mkidx, or other problem.
		// Assume index name + .mkv extension
		
		ret = ReplaceExtension(szIndex, L".mkv");
	}
	
	return ret;
}


wchar_t *InputFileMKV::GetIndexName(const wchar_t *szFile)
{
	return ReplaceExtension(szFile, L".mkidx");
}


bool InputFileMKV::ReadIndex(const wchar_t *szFile)
{
	WinFile  idx;
	DWORD    ver;
	int      tk;
	__int64  tmpTimeScale;
	int      tmpTracks;
	int      tmpAudioTracks;
	int      tmpVideoTracks;

	bool ret = false;

	wchar_t *szIndex = GetIndexName(szFile);
	if (szIndex == NULL) goto Abort;

	if (!idx.Open(szIndex, false)) goto Abort;
	delete[] szIndex;
	szIndex = NULL;

	// Version
	idx.Read(&ver, sizeof(ver));
	if (ver != 'DIKM') goto Abort;
	idx.Read(&ver, sizeof(ver));

	// ........ ........ ........ 01011000 'X' (0x58)
	// ........ ........ 00110000 ........ '0' (0x30)
	// ........ 00110000 ........ ........ '0' (0x30)
	// 00110001 ........ ........ ........ '1' (0x31) or
	// 00110010 ........ ........ ........ '2' (0x32) or
	// 00110011 ........ ........ ........ '3' (0x33) or
	// 00110100 ........ ........ ........ '4' (0x34)

	if ((ver & 0xF8FFFFFF) != '000X') goto Abort;

	// Filename (version >= 2), ignore
	if (ver >= '200X')
	{
		idx.Read(&tk, sizeof(tk));
		idx.Move(tk * sizeof(wchar_t));
	}

	// InputFileMKV data
	idx.Read(&tmpTimeScale,   sizeof(tmpTimeScale));
	idx.Read(&tmpTracks,      sizeof(tmpTracks));
	idx.Read(&tmpAudioTracks, sizeof(tmpAudioTracks));
	idx.Read(&tmpVideoTracks, sizeof(tmpVideoTracks));

	// MatroskaTrack data
	iTracks = 0;

	for (tk = 0; tk < tmpTracks; ++tk)
	{
		MatroskaTrack *mt = new MatroskaTrack;
		if (mt == NULL) break;

		if (!mt->ReadFromIndex(idx, ver))
		{
			delete mt;
			break;
		}

		Track[iTracks++] = mt;
	}

	ret = (iTracks > 0);
	if (ret)
	{
		iTimecodeScale = tmpTimeScale;
		iAudioTracks   = tmpAudioTracks;
		iVideoTracks   = tmpVideoTracks;
	}

Abort:
	delete[] szIndex;
	return ret;
}


bool InputFileMKV::WriteIndex(const wchar_t *szFile)
{
	WinFile        idx;
	DWORD          ver;
	int            tk;
	wchar_t*       szIndex;
	const wchar_t* szTitle;

	bool ret = false;

	szIndex = GetIndexName(szFile);
	if (szIndex == NULL) goto Abort;

	if (!idx.Open(szIndex, true)) goto Abort;
	delete[] szIndex;
	szIndex = NULL;

	// Version
	ver = 'DIKM';
	idx.Write(&ver, sizeof(ver));
	ver = '400X';
	idx.Write(&ver, sizeof(ver));

	// File Title
	szTitle = GetFileTitle(szFile);
	tk = (szTitle != NULL)? lstrlenW(szTitle): 0;
	idx.Write(&tk, sizeof(tk));
	if (tk > 0) idx.Write(szTitle, tk * sizeof(wchar_t));

	// InputFileMKV data
	idx.Write(&iTimecodeScale, sizeof(iTimecodeScale));
	idx.Write(&iTracks,        sizeof(iTracks));
	idx.Write(&iAudioTracks,   sizeof(iAudioTracks));
	idx.Write(&iVideoTracks,   sizeof(iVideoTracks));

	// MatroskaTrack data
	for (tk = 0; tk < iTracks; ++tk)
	{
		if (!Track[tk]->WriteToIndex(idx)) goto Abort;
	}

	ret = true;

Abort:
	delete[] szIndex;
	return ret;
}

