///////////////////////////////////////////////////////////////////////
// Matroska plugin for VirtualDub 1.8.8+
// Copyright (C) 2010-2012 fccHandler
// Copyright (C) 1998-2012 Avery Lee
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////

// IVDXVideoDecoder, IVDXVideoDecoderModel,
// and IVDXVideoSource implementations

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <vfw.h>
#include <crtdbg.h>

// Use VirtualDub's I/P decoder model?
#define USE_DEFAULT_IP_MODEL

// Defer video decoding to VirtualDub?
//#define DEFER_DECODING

// Unwrap AVC frames in VideoDecoderMKV?
// (No, this breaks direct stream copy of AVC.)
//#define DECODER_UNWRAP_AVC


#include "VDInputDriver.h"
#include "VDUnknown.h"
#include "MkvFile.h"

#include "DataVector.h"
#include "VideoCodec.h"

#include "InputFileMKV.h"
#include "MatroskaTrack.h"
#include "VideoSourceMKV.h"


///////////////////////////////////////////////////////////////////////
//
//                     VideoDecoderMKV
//
///////////////////////////////////////////////////////////////////////

class VideoDecoderMKV : public vdxunknown<IVDXVideoDecoder> {

public:
	VideoDecoderMKV(VideoSourceMKV* const pp, MatroskaTrack* const mt);
	~VideoDecoderMKV();

	// IVDXVideoDecoder
	const void*      VDXAPIENTRY DecodeFrame(const void *inputBuffer, uint32 data_len, bool is_preroll, sint64 sampleNumber, sint64 targetFrame);
	uint32           VDXAPIENTRY GetDecodePadding();
	void             VDXAPIENTRY Reset();
	bool             VDXAPIENTRY IsFrameBufferValid();
	const VDXPixmap& VDXAPIENTRY GetFrameBuffer();
	bool             VDXAPIENTRY SetTargetFormat(int format, bool useDIBAlignment);
	bool             VDXAPIENTRY SetDecompressedFormat(const VDXBITMAPINFOHEADER *pbih);
	bool             VDXAPIENTRY IsDecodable(sint64 sample_num);
	const void*      VDXAPIENTRY GetFrameBufferBase();

private:
	VideoSourceMKV*  const parentPtr;
	MatroskaTrack*   const video;

	VideoCodec       mCodec;

	sint64           mPrevFrame;
	unsigned long    dwActualFrame;

#ifdef DECODER_UNWRAP_AVC
	unsigned char*   avc_buf;
	unsigned long    avc_bufsize;

	int UnwrapAVCFrame(unsigned char *dst, const unsigned char *src, int len);
#endif
};


VideoDecoderMKV::VideoDecoderMKV(VideoSourceMKV *const pp, MatroskaTrack *const mt)
	: parentPtr     (pp)
	, video         (mt)
#ifdef DECODER_UNWRAP_AVC
	, avc_buf       (NULL)
	, avc_bufsize   (0)
#endif
{
	parentPtr->AddRef();

	if (video != NULL)
	{
		video->hCodec = NULL;

		dwActualFrame = 0xFFFFFFFF;
		mPrevFrame    = -1;
	}
	else
	{
		mPrevFrame    = 0;
	}
}


VideoDecoderMKV::~VideoDecoderMKV()
{
	if (video != NULL)
	{
		video->hCodec = NULL;
#ifdef DECODER_UNWRAP_AVC
		delete[] avc_buf;
#endif
	}
	parentPtr->Release();
}


#ifdef DECODER_UNWRAP_AVC

int VideoDecoderMKV::UnwrapAVCFrame(unsigned char *dst, const unsigned char *src, int len)
{
	// Unwrap NALUs
	int avc_len = video->avc_len;
	int ret = 0;

	while (len > avc_len)
	{
		int size = *src++;
		--len;
		if (avc_len > 1) {
			size = (size << 8) + *src++;
			--len;
			if (avc_len > 2) {
				size = (size << 8) + *src++;
				--len;
				if (avc_len > 3) {
					size = (size << 8) + *src++;
					--len;
				}
			}
		}

		if (size < 0) break;

		if (size > 0)
		{
			if (size > len) size = len;

			if (dst != NULL)
			{
				*dst++ = 0;
				*dst++ = 0;
				*dst++ = 0;
				*dst++ = 1;
				memcpy(dst, src, size);
				dst += size;
			}

			ret += size + 4;
			src += size;
			len -= size;
		}
	}

	return ret;
}

#endif  // DECODER_UNWRAP_AVC


const void *VideoDecoderMKV::DecodeFrame(
	const void *inputBuffer, uint32 data_len, bool is_preroll,
	sint64 streamFrame,	sint64 targetFrame)
{
	if (video != NULL)
	{
		DWORD dwFrame;
	
		if (streamFrame < 0) streamFrame = targetFrame;

		dwFrame = (DWORD)streamFrame;

		// If dwFrame is the same as dwActualFrame,
		// then we've already decoded this frame.

		if (dwFrame < video->dwActualLength && dwFrame != dwActualFrame)
		{
			DWORD f = video->sample_list[dwFrame].size;

			dwActualFrame = dwFrame;

			if (f & MKV_KEYFRAME) {
				f = 0;
			} else {
				f = ICDECOMPRESS_NOTKEYFRAME;
			}

			if (is_preroll) {
				f |= ICDECOMPRESS_PREROLL;
			} else {
				mPrevFrame = streamFrame;
			}

#ifdef DECODER_UNWRAP_AVC
			if (video->iCodecID == V_MPEG4_ISO_AVC)
			{
				const unsigned char *src;
				DWORD dwPreLen;
				DWORD dwSrcLen;
				DWORD dwDstLen;

				// The data we receive contains the following:
				//  - private data, if it is a keyframe
				//  - comp settings, if header compression is used
				//  - the raw AVC data (wrapped)
				// We need to unwrap the raw AVC data for ffdshow.

				// Determine the destination buffer size we will
				// need and get a pointer to the source AVC data.

				dwPreLen = (f & ICDECOMPRESS_NOTKEYFRAME)? 0: video->iCodecPrivateLen;

				src = (unsigned char *)inputBuffer + dwPreLen;
				dwSrcLen = data_len - dwPreLen;
				dwDstLen = dwPreLen + UnwrapAVCFrame(NULL, src, dwSrcLen);

				// Grow the destination buffer if necessary
				if (dwDstLen > avc_bufsize)
				{
					delete[] avc_buf;
					avc_buf = new unsigned char[dwDstLen];
					if (avc_buf == NULL) return mCodec.mFrameBuffer;
					avc_bufsize = dwDstLen;
				}

				// Now unwrap the frame.
				if (dwPreLen > 0)
				{
					memcpy(avc_buf, inputBuffer, dwPreLen);
				}
				UnwrapAVCFrame(avc_buf + dwPreLen, src, dwSrcLen);

				// Replace the parameters for the Decompress call.
				inputBuffer = avc_buf;
				data_len = dwDstLen;
			}
#endif  // DECODER_UNWRAP_AVC

			mCodec.Decompress((LPVOID)inputBuffer, data_len, f);
		}
	}

	return mCodec.mFrameBuffer;
}


uint32 VideoDecoderMKV::GetDecodePadding()
{
	return 0;
}


void VideoDecoderMKV::Reset()
{
	if (video != NULL)
	{
		dwActualFrame = 0xFFFFFFFF;
		mPrevFrame    = -1;
	}
}


bool VideoDecoderMKV::IsFrameBufferValid()
{
	return (mPrevFrame >= 0);
}


const VDXPixmap& VideoDecoderMKV::GetFrameBuffer()
{
	return mCodec.mPixmap;
}


bool VideoDecoderMKV::SetTargetFormat(int format, bool useDIBAlignment)
{
	if (video != NULL)
	{
		Reset();

		if (mCodec.Init((BITMAPINFOHEADER *)video->pFormat, format)) {
			video->hCodec = mCodec.hCodec;
			return true;
		}

		if (mCodec.hCodec != NULL) return false;
		video->hCodec = NULL;

		// Missing codec
		mCodec.Inform("MISSING CODEC");

		return (format == mCodec.mPixmap.format
		    ||  format == nsVDXPixmap::kPixFormat_Null);
	}

	// Emulated video source

	if (mCodec.Init(NULL, format))
	{
		// No video stream
		mCodec.Inform("NO VIDEO");

		return (format == mCodec.mPixmap.format
		     || format == nsVDXPixmap::kPixFormat_Null);
	}

	return false;
}


bool VideoDecoderMKV::SetDecompressedFormat(const VDXBITMAPINFOHEADER *pbih)
{
	return false;
}


const void *VideoDecoderMKV::GetFrameBufferBase()
{
	return mCodec.mFrameBuffer;
}


bool VideoDecoderMKV::IsDecodable(sint64 sample_num64)
{
	if (video != NULL)
	{
		if ((DWORD)sample_num64 < video->dwActualLength)
		{
			return (sample_num64 == (mPrevFrame + 1));
		}
		return false;
	}
	return true;
}


#ifndef USE_DEFAULT_IP_MODEL

///////////////////////////////////////////////////////////////////////
//
//                      VideoDecoderModelMKV
//
///////////////////////////////////////////////////////////////////////

class VideoDecoderModelMKV : public vdxunknown<IVDXVideoDecoderModel> {

public:
	VideoDecoderModelMKV(VideoSourceMKV *const pp, MatroskaTrack *const mt);
	~VideoDecoderModelMKV();

	// IVDXVideoDecoderModel
	void    VDXAPIENTRY Reset();
	void    VDXAPIENTRY SetDesiredFrame(sint64 frame_num);
	sint64  VDXAPIENTRY GetNextRequiredSample(bool& is_preroll);
	int     VDXAPIENTRY GetRequiredCount();

private:
	VideoSourceMKV *const parentPtr;
	MatroskaTrack  *const video;

	sint64  mLastFrame;
	sint64  mTargetFrame;
	sint64  mNextFrame;

	bool    IsKey(sint64 sample);
};


VideoDecoderModelMKV::VideoDecoderModelMKV(VideoSourceMKV *const pp, MatroskaTrack *const mt)
	: parentPtr(pp)
	, video(mt)
{
	parentPtr->AddRef();

	mLastFrame = -1;
}


VideoDecoderModelMKV::~VideoDecoderModelMKV()
{
	parentPtr->Release();
}


void VideoDecoderModelMKV::Reset()
{
	mLastFrame = -1;
}


bool VideoDecoderModelMKV::IsKey(sint64 sample)
{
	DWORD dwActual = (DWORD)sample;
	if (dwActual < video->dwActualLength)
	{
		return ((video->sample_list[dwActual].size & MKV_KEYFRAME) != 0);
	}
	return false;
}


void VideoDecoderModelMKV::SetDesiredFrame(sint64 frame_num)
{
	mTargetFrame = frame_num;

	// Do we already have the frame?
	if (mLastFrame == frame_num) {
		mNextFrame = -1;	// Flag this for GetNextRequiredSample
		return;
	}

	mNextFrame = frame_num;

	// If it's a Delta frame, locate a previous Key
	if (!IsKey(frame_num)) {

		while (mNextFrame > 0) {
			--mNextFrame;
			if (mNextFrame == mLastFrame) {
				// Start from the frame following that one
				++mNextFrame;
				break;
			}
			if (IsKey(mNextFrame)) break;
		}
	}
}


sint64 VideoDecoderModelMKV::GetNextRequiredSample(bool& is_preroll)
{
	sint64 frame;

	if (mNextFrame == mTargetFrame || mNextFrame < 0) {
		is_preroll = false;
		mLastFrame = mTargetFrame;
		frame = mNextFrame;
		mNextFrame = -1;
	} else {
		is_preroll = true;
		frame = mNextFrame++;
	}

	return frame;
}


int VideoDecoderModelMKV::GetRequiredCount()
{
	if (mNextFrame < 0) return 0;

	return (int)((mTargetFrame - mNextFrame) + 1);
}

#endif	// USE_DEFAULT_IP_MODEL


///////////////////////////////////////////////////////////////////////
//
//                      VideoSourceMKV
//
///////////////////////////////////////////////////////////////////////


VideoSourceMKV::VideoSourceMKV(InputFileMKV *const pp, MatroskaTrack *const vt, MatroskaTrack *const at)
	: parentPtr(pp)
	, video(vt)
	, audio(at)
{
	parentPtr->AddRef();

	errmode         = kErrorModeDecodeAnyway;
	dwPrevSample    = 0xFFFFFFFF;

	if (video != NULL)
	{
		unsigned long i;
		unsigned long max = 0;
		
		for (i = 0; i < video->dwActualLength; ++i)
		{
			unsigned long sz = video->sample_list[i].size & (~MKV_KEYFRAME);
			if (sz > max) max = sz;
		}
		max += video->iCompSettingsLen;

		if (max < 4) max = 4;
		frame_buffer = new unsigned char[max];
	}
	else
	{
		// Emulated video source
		frame_buffer = NULL;
	}
}


VideoSourceMKV::~VideoSourceMKV()
{
	delete[] frame_buffer;

	parentPtr->Release();
}


int VideoSourceMKV::AddRef()
{
	return vdxunknown<IVDXStreamSource>::AddRef();
}


int VideoSourceMKV::Release()
{
	return vdxunknown<IVDXStreamSource>::Release();
}


void *VDXAPIENTRY VideoSourceMKV::AsInterface(uint32 iid)
{
	if (iid == IVDXVideoSource::kIID)
		return static_cast<IVDXVideoSource *>(this);

	return vdxunknown<IVDXStreamSource>::AsInterface(iid);
}


void VDXAPIENTRY VideoSourceMKV::GetStreamSourceInfo(VDXStreamSourceInfo& srcInfo)
{
	if (video != NULL)
	{
		srcInfo.mSampleRate.mNumerator   = video->dwRate;
		srcInfo.mSampleRate.mDenominator = video->dwScale;
		srcInfo.mSampleCount             = video->dwActualLength;
	}
	else
	{
		// Emulated video source
		const WAVEFORMATEX *wfex = (WAVEFORMATEX *)audio->pFormat;

		srcInfo.mSampleRate.mNumerator   = wfex->nSamplesPerSec;
		srcInfo.mSampleRate.mDenominator = 1000;

		srcInfo.mSampleCount = (sint64)(
			(double)wfex->nSamplesPerSec * audio->Duration / 1000.0 + 0.5
		);

		// Err on the side of caution to avoid cutting off any audio.
		++srcInfo.mSampleCount;
	}

	// TODO:
	srcInfo.mPixelAspectRatio.mNumerator   = 1;
	srcInfo.mPixelAspectRatio.mDenominator = 1;
}


int VideoSourceMKV::UnwrapAVCFrame(unsigned char *dst, const unsigned char *src, int len)
{
	// Unwrap NALUs
	int avc_len = video->avc_len;
	int ret = 0;

	while (len > avc_len)
	{
		int size = *src++;
		--len;
		if (avc_len > 1) {
			size = (size << 8) + *src++;
			--len;
			if (avc_len > 2) {
				size = (size << 8) + *src++;
				--len;
				if (avc_len > 3) {
					size = (size << 8) + *src++;
					--len;
				}
			}
		}

		if (size < 0) break;

		if (size > 0)
		{
			if (size > len) size = len;

			if (dst != NULL)
			{
				*dst++ = 0;
				*dst++ = 0;
				*dst++ = 0;
				*dst++ = 1;
				memcpy(dst, src, size);
				dst += size;
			}

			ret += size + 4;
			src += size;
			len -= size;
		}
	}

	return ret;
}


bool VideoSourceMKV::Read(sint64 lStart64, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lBytesRead, uint32 *lSamplesRead)
{
	unsigned char *dst = (unsigned char *)lpBuffer;
	DWORD framesize;
	DWORD actual;
	DWORD dwStart;

	if (video == NULL)
	{
		// Emulated video source
		if (lBytesRead)   *lBytesRead   = 4096;
		if (lSamplesRead) *lSamplesRead = 1;

		return (lpBuffer == NULL || cbBuffer >= 4096);
	}

	dwStart = (DWORD)lStart64;

	if (dwStart >= video->dwActualLength)
	{
		if (lBytesRead) *lBytesRead = 0;
		if (lSamplesRead) *lSamplesRead = 0;
		return true;
	}

	framesize = video->sample_list[dwStart].size & (~MKV_KEYFRAME);

_RPT2(_CRT_WARN, "VideoSourceMKV::Read %lu, %lu\n", dwStart, dwPrevSample);

	if (video->iCodecID == V_MPEG4_ISO_AVC)
	{
#ifdef DECODER_UNWRAP_AVC
		actual = video->iCompSettingsLen + framesize;
#else
		// read data, but only if it's new data
		if (dwStart != dwPrevSample)
		{
			dwPrevSample = dwStart;
			if (video->iCompSettingsLen > 0)
			{
				memcpy(
					frame_buffer,
					video->pCompSettings,
					video->iCompSettingsLen
				);
			}
			parentPtr->ReadData(
				frame_buffer + video->iCompSettingsLen,
				video->sample_list[dwStart].pos,
				framesize
			);
		}
		actual = UnwrapAVCFrame(
			NULL,
			frame_buffer,
			video->iCompSettingsLen + framesize
		);
#endif
		if (video->sample_list[dwStart].size & MKV_KEYFRAME)
		{
			actual += video->iCodecPrivateLen;
		}
	}
	else
	{
		// Not AVC
		actual = video->iCompSettingsLen + framesize;
	}

	if (dst == NULL || cbBuffer < actual)
	{
		if (lBytesRead) *lBytesRead = actual;
		if (lSamplesRead) *lSamplesRead = 1;
		return (dst == NULL);
	}

	if (video->iCodecID == V_MPEG4_ISO_AVC)
	{
		if (video->sample_list[dwStart].size & MKV_KEYFRAME)
		{
			if (video->iCodecPrivateLen > 0)
			{
				memcpy(dst, video->pCodecPrivate, video->iCodecPrivateLen);
				dst += video->iCodecPrivateLen;
			}
		}
#ifdef DECODER_UNWRAP_AVC
		if (video->iCompSettingsLen > 0)
		{
			memcpy(dst, video->pCompSettings, video->iCompSettingsLen);
			dst += video->iCompSettingsLen;
		}
		parentPtr->ReadData(dst,
			video->sample_list[dwStart].pos,
			framesize
		);
		dst += framesize;
#else
		dst += UnwrapAVCFrame(
			dst,
			frame_buffer,
			video->iCompSettingsLen + framesize
		);
#endif
	}
	else
	{
		// Not AVC
		// read data, but only if it's new data
		if (dwStart != dwPrevSample)
		{
			dwPrevSample = dwStart;
			if (video->iCompSettingsLen > 0)
			{
				memcpy(
					frame_buffer,
					video->pCompSettings,
					video->iCompSettingsLen
				);
			}
			parentPtr->ReadData(
				frame_buffer + video->iCompSettingsLen,
				video->sample_list[dwStart].pos,
				framesize
			);
		}
		memcpy(dst, frame_buffer, actual);
		dst += actual;
	}

	if (lBytesRead)
	{
		*lBytesRead = (uint32)(dst - (unsigned char *)lpBuffer);
	}

	if (lSamplesRead) *lSamplesRead = 1;

	return true;
}


const void *VideoSourceMKV::GetDirectFormat()
{
	if (video != NULL) return video->pFormat;
	return NULL;
}


int VideoSourceMKV::GetDirectFormatLen()
{
	if (video != NULL) return video->iFormatLen;
	return 0;
}


IVDXStreamSource::ErrorMode VideoSourceMKV::GetDecodeErrorMode()
{
	return errmode;
}


void VideoSourceMKV::SetDecodeErrorMode(IVDXStreamSource::ErrorMode mode)
{
	errmode = mode;
}


bool VideoSourceMKV::IsDecodeErrorModeSupported(IVDXStreamSource::ErrorMode mode)
{
	return true;
}


bool VideoSourceMKV::IsVBR()
{
	return false;
}


sint64 VideoSourceMKV::TimeToPositionVBR(sint64 us)
{
	return 0;
}


sint64 VideoSourceMKV::PositionToTimeVBR(sint64 samples)
{
	return 0;
}


void VideoSourceMKV::GetVideoSourceInfo(VDXVideoSourceInfo& info)
{
	if (video != NULL)
	{
		const BITMAPINFOHEADER *bmih = (BITMAPINFOHEADER *)video->pFormat;

		info.mWidth  = bmih->biWidth;
		info.mHeight = bmih->biHeight;
		info.mFlags  = 0;

#ifdef USE_DEFAULT_IP_MODEL
		info.mDecoderModel = VDXVideoSourceInfo::kDecoderModelDefaultIP;
#else
		info.mDecoderModel = VDXVideoSourceInfo::kDecoderModelCustom;
#endif
	}
	else
	{
		// Emulated video source
		info.mWidth  = 320;
		info.mHeight = 240;
		info.mFlags  = 0;

		info.mDecoderModel = VDXVideoSourceInfo::kDecoderModelDefaultIP;
	}
}


bool VideoSourceMKV::CreateVideoDecoderModel(IVDXVideoDecoderModel **ppModel)
{
#ifndef USE_DEFAULT_IP_MODEL
	if (video != NULL)
	{
		*ppModel = new VideoDecoderModelMKV(this, video);
		if (*ppModel == NULL) return false;

		(*ppModel)->AddRef();
		return true;
	}
#endif
	*ppModel = NULL;
	return false;
}


bool VideoSourceMKV::CreateVideoDecoder(IVDXVideoDecoder **ppDecoder)
{
#ifdef DEFER_DECODING
	*ppDecoder = NULL;
	return false;
#else
	*ppDecoder = new VideoDecoderMKV(this, video);
	if (*ppDecoder == NULL) return false;

	(*ppDecoder)->AddRef();
	return true;
#endif
}


void VideoSourceMKV::GetSampleInfo(sint64 sample_num, VDXVideoFrameInfo& frameInfo)
{
	if (video != NULL)
	{
		DWORD dwActual = (DWORD)sample_num;

		if (dwActual < video->dwActualLength)
		{
			const MKVSampleInfo *msi = &(video->sample_list[dwActual]);

			frameInfo.mBytePosition = msi->pos;

			if (msi->size & MKV_KEYFRAME) {
				frameInfo.mFrameType = kVDXVFT_Independent;
				frameInfo.mTypeChar  = 'K';
			} else {
				frameInfo.mFrameType = kVDXVFT_Predicted;
				frameInfo.mTypeChar  = ' ';
			}

		} else {
			frameInfo.mBytePosition = -1;
			frameInfo.mFrameType    = kVDXVFT_Null;
			frameInfo.mTypeChar     = 'M';
		}
	}
	else
	{
		// Emulated video source
		if (IsKey(sample_num))
		{
			frameInfo.mTypeChar  = 'K';
			frameInfo.mFrameType = kVDXVFT_Independent;
		}
		else
		{
			frameInfo.mTypeChar  = ' ';
			frameInfo.mFrameType = kVDXVFT_Predicted;
		}
		frameInfo.mBytePosition  = -1;
	}
}


bool VideoSourceMKV::IsKey(sint64 sample)
{
	if (video != NULL)
	{
		DWORD dwActual = (DWORD)sample;
		if (dwActual < video->dwActualLength)
		{
			return ((video->sample_list[dwActual].size & MKV_KEYFRAME) != 0);
		}
		return false;
	}

	// Emulated video source
	const WAVEFORMATEX *wfex = (WAVEFORMATEX *)audio->pFormat;

	return (((sample * 100) % wfex->nSamplesPerSec) == 0);
}


sint64 VideoSourceMKV::GetFrameNumberForSample(sint64 sample_num)
{
	return sample_num;
}


sint64 VideoSourceMKV::GetSampleNumberForFrame(sint64 display_num)
{
	return display_num;
}


sint64 VideoSourceMKV::GetRealFrame(sint64 display_num)
{
	return display_num;
}


sint64 VideoSourceMKV::GetSampleBytePosition(sint64 sample_num)
{
	if (video != NULL)
	{
		DWORD dwActual = (DWORD)sample_num;
		if (dwActual < video->dwActualLength)
		{
			return video->sample_list[dwActual].pos;
		}
	}
	return -1;
}

