///////////////////////////////////////////////////////////////////////
// Matroska plugin for VirtualDub 1.8.8+
// Copyright (C) 2010-2012 fccHandler
// Copyright (C) 1998-2012 Avery Lee
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////
//
//	Windows file I/O functions, wrapped in a class.
//	Files larger than 4GB are supported on NTFS.
//	Any file I/O error disables all subsequent file functions.
//
///////////////////////////////////////////////////////////////////////

#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include "WinFile.h"


WinFile::WinFile()
{
	// Constructor

	m_Handle = INVALID_HANDLE_VALUE;
	Kill();
}


WinFile::~WinFile()
{
	// Destructor

	Kill();
}


bool WinFile::Kill()
{
	// Close the file handle if applicable and clear all state.

	bool ret = true;

	if (m_Handle != INVALID_HANDLE_VALUE)
	{
		if (m_Flags & WF_WRITE)
		{
			if (m_Flags & WF_DIRTY)
			{
				ret = Flush();
				if (!ret) goto Abort;
			}

			m_Size.LowPart = SetFilePointer(m_Handle, m_Size.LowPart, &m_Size.HighPart, FILE_BEGIN);

			if (m_Size.LowPart != INVALID_SET_FILE_POINTER || GetLastError() == NO_ERROR)
			{
				SetEndOfFile(m_Handle);
			}
			else
			{
				ret = false;
			}
		}

Abort:
		if (m_Flags & WF_CLOSE)
		{
			if (CloseHandle(m_Handle) == FALSE) ret = false;
		}

		m_Handle = INVALID_HANDLE_VALUE;
	}

	ZeroMemory(m_Buff, WF_BUFFERSIZE);

	m_BuffPos.QuadPart  = 0;
	m_Size.QuadPart     = 0;
	m_Pos.QuadPart      = 0;
	m_Flags             = 0;
	m_CurBit            = 0;
	m_CurByte           = 0;

	return ret;
}


/********************************************************************
 *
 *	How Windows handles some aberrant situations:
 *
 *	Setting the file pointer to absolute -10 returns 0xFFFFFFFF
 *	and error code 87 (ERROR_INVALID_PARAMETER), and does
 *	not move the file pointer (apparently)
 *
 *	Setting the file pointer to relative -10 returns 0xFFFFFFFF
 *	and error code 131 (ERROR_NEGATIVE_SEEK), and does
 *	not move the file pointer
 *
 *	Setting the file pointer AT or BEYOND the end of file succeeds.
 *
 *	Attempting to read past the end of file does not cause
 *	an error, but reads zero bytes.
 *
 ********************************************************************/

bool WinFile::Flush()
{
	// Flush from current buffer position

	LONG  lHi  = m_BuffPos.HighPart;
	DWORD dwLo = m_BuffPos.LowPart & (~(WF_BUFFERSIZE - 1));

	dwLo = SetFilePointer(m_Handle, dwLo, &lHi, FILE_BEGIN);

	if (dwLo == INVALID_SET_FILE_POINTER)
	{
		if (GetLastError() != NO_ERROR) goto Abort;
	}

	if (WriteFile(m_Handle, m_Buff, WF_BUFFERSIZE, &dwLo, NULL) != FALSE)
	{
		return true;
	}

Abort:
	return false;
}


bool WinFile::Refresh()
{
	// Refresh from current file position

	LONG  lHi  = m_Pos.HighPart;
	DWORD dwLo = m_Pos.LowPart & (~(WF_BUFFERSIZE - 1));

	dwLo = SetFilePointer(m_Handle, dwLo, &lHi, FILE_BEGIN);

	if (dwLo == INVALID_SET_FILE_POINTER)
	{
		if (GetLastError() != NO_ERROR) goto Abort;
	}

	m_BuffPos.LowPart  = dwLo;
	m_BuffPos.HighPart = lHi;

	if (ReadFile(m_Handle, m_Buff, WF_BUFFERSIZE, &dwLo, NULL) != FALSE)
	{
		if (dwLo < WF_BUFFERSIZE)
		{
			ZeroMemory(m_Buff + dwLo, WF_BUFFERSIZE - dwLo);
		}
		return true;
	}

Abort:
	return false;
}


// static
LPSTR WinFile::WcharToAnsi(LPCWSTR lpszWchar)
{
	// Convert a wide character string to ANSI.

	if (lpszWchar != NULL)
	{
		int len = WideCharToMultiByte(CP_ACP, 0, lpszWchar, -1, NULL, 0, NULL, NULL);
		if (len > 0)
		{
			char *ret = new char[len + 2];
			if (ret != NULL)
			{
				if (WideCharToMultiByte(CP_ACP, 0, lpszWchar, -1, ret, len + 1, NULL, NULL) > 0)
				{
					ret[len + 1] = '\0';
					return ret;
				}
				delete[] ret;
			}
		}
	}
	return NULL;
}


// static
bool WinFile::Delete(const char *lpszFileName)
{
	// Delete a file, ANSI version

	if (lpszFileName != NULL)
	{
		return (DeleteFileA(lpszFileName) != FALSE);
	}
	return false;
}


// static
bool WinFile::Delete(const wchar_t *lpszFileName)
{
	// Delete a file, wide character version

	bool ret = false;

	if (lpszFileName != NULL)
	{
		if (GetVersion() < 0x80000000U)
		{
			// Windows NT/XP
			ret = (DeleteFileW(lpszFileName) != FALSE);
		}
		else
		{
			// Windows 95/98
			LPSTR tmp = WcharToAnsi(lpszFileName);

			if (tmp != NULL)
			{
				ret = (DeleteFileA(tmp) != FALSE);
				delete[] tmp;
			}
		}
	}
	
	return ret;
}


HANDLE WinFile::DualCreateFile(
	LPCWSTR                 lpFileName,
	DWORD                   dwDesiredAccess,
	DWORD                   dwShareMode,
	LPSECURITY_ATTRIBUTES   lpSecurityAttributes,
	DWORD                   dwCreationDisposition,
	DWORD                   dwFlagsAndAttributes,
	HANDLE                  hTemplateFile
	) const
{
	// Wrapper for CreateFileW and CreateFileA.

	HANDLE h = INVALID_HANDLE_VALUE;

	if (GetVersion() < 0x80000000U)
	{
		// Windows NT/XP
		h = CreateFileW(lpFileName, dwDesiredAccess, dwShareMode,
			lpSecurityAttributes, dwCreationDisposition,
			dwFlagsAndAttributes, hTemplateFile);
	}
	else
	{
		// Windows 95/98
		LPSTR tmp = WcharToAnsi(lpFileName);

		if (tmp != NULL)
		{
			h = CreateFileA(tmp, dwDesiredAccess, dwShareMode,
				lpSecurityAttributes, dwCreationDisposition,
				dwFlagsAndAttributes, hTemplateFile);

			delete[] tmp;
		}
	}

	return h;
}


bool WinFile::Bind(HANDLE h, bool bWriteAccess)
{
	// Bind our state to an existing file handle.

	LARGE_INTEGER size;
	LARGE_INTEGER ptr;

	bool ret = false;

	if (h == INVALID_HANDLE_VALUE) goto Abort;

	// Get the file size

	size.HighPart = 0;
	size.LowPart = GetFileSize(h, (LPDWORD)&size.HighPart);

	if (size.LowPart == INVALID_FILE_SIZE)
	{
		if (GetLastError() != NO_ERROR) goto Abort;
	}
	if (size.HighPart < 0) goto Abort;

	// Get the current file pointer

	ptr.HighPart = 0;
	ptr.LowPart  = SetFilePointer(h, 0, &ptr.HighPart, FILE_CURRENT);

	if (ptr.LowPart == INVALID_SET_FILE_POINTER)
	{
		if (GetLastError() != NO_ERROR) goto Abort;
	}
	if (ptr.HighPart < 0) goto Abort;

	// Bind to the file

	Kill();

	m_Handle        = h;
	m_Pos.QuadPart  = ptr.QuadPart;
	m_Size.QuadPart = size.QuadPart;

	if (bWriteAccess) m_Flags = WF_WRITE;

	ret = Refresh();

	// If we fail, kill us but do not close the handle

	if (!ret)
	{
		m_Handle = INVALID_HANDLE_VALUE;
		Kill();
	}

Abort:
	return ret;
}


bool WinFile::Open(const char *lpszFileName, bool bWriteAccess)
{
	// Open a file for read (and possibly write) access.
	// ANSI version.

	HANDLE h;

	if (lpszFileName == NULL) goto Abort;

	h = CreateFileA(
		(LPCSTR)lpszFileName,
		bWriteAccess? GENERIC_READ | GENERIC_WRITE: GENERIC_READ,
		FILE_SHARE_READ,
		NULL,
		OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL,
		NULL
	);

	if (h == INVALID_HANDLE_VALUE)
	{
		if (!bWriteAccess) goto Abort;

		// Attempt to create a new file
		h = CreateFileA(
			(LPCSTR)lpszFileName,
			GENERIC_READ | GENERIC_WRITE,
			FILE_SHARE_READ,
			NULL,
			CREATE_ALWAYS,
			FILE_ATTRIBUTE_NORMAL,
			NULL
		);

		if (h == INVALID_HANDLE_VALUE) goto Abort;
	}

	if (Bind(h, bWriteAccess))
	{
		m_Flags |= WF_CLOSE;
		return true;
	}

	CloseHandle(h);

Abort:
	return false;
}


bool WinFile::Open(const wchar_t *lpszFileName, bool bWriteAccess)
{
	// Open a file for read (and possibly write) access.
	// Wide character version.

	HANDLE h;

	if (lpszFileName == NULL) goto Abort;

	h = DualCreateFile(
		(LPCWSTR)lpszFileName,
		bWriteAccess? GENERIC_READ | GENERIC_WRITE: GENERIC_READ,
		FILE_SHARE_READ,
		NULL,
		OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL,
		NULL
	);

	if (h == INVALID_HANDLE_VALUE)
	{
		if (!bWriteAccess) goto Abort;

		// Attempt to create a new file
		h = DualCreateFile(
			(LPCWSTR)lpszFileName,
			GENERIC_READ | GENERIC_WRITE,
			FILE_SHARE_READ,
			NULL,
			CREATE_ALWAYS,
			FILE_ATTRIBUTE_NORMAL,
			NULL
		);

		if (h == INVALID_HANDLE_VALUE) goto Abort;
	}

	if (Bind(h, bWriteAccess))
	{
		m_Flags |= WF_CLOSE;
		return true;
	}

	CloseHandle(h);

Abort:
	return false;
}


bool WinFile::Close()
{
	// Close the file unconditionally and clear all state.

	m_Flags |= WF_CLOSE;
	return Kill();
}


bool WinFile::Seek(__int64 iAbsPos)
{
	// Seek to an absolute position in the file.

	bool ret = false;

	if (m_Handle == INVALID_HANDLE_VALUE) goto Abort;
	if (iAbsPos < 0) goto Abort;

	if (iAbsPos > m_Size.QuadPart)
	{
		// Do not allow seeking beyond EOF
		iAbsPos = m_Size.QuadPart;
	}

	// Is the new position within the same file page?
	ret = (
		(m_Pos.QuadPart & -((LONGLONG)WF_BUFFERSIZE) )
		== (    iAbsPos & -((LONGLONG)WF_BUFFERSIZE) )
	);

	m_Pos.QuadPart = iAbsPos;

	if (!ret) {
		// New position is in a different file page.

		ret = ((m_Flags & WF_DIRTY) == 0);

		if (!ret)
		{
			m_Flags &= (~WF_DIRTY);
			ret = Flush();
		}

		if (ret) ret = Refresh();

		if (!ret) Kill();
	}

Abort:
	return ret;
}


bool WinFile::Move(int iDistance)
{
	// Move a relative distance from the current position.

	return Seek(m_Pos.QuadPart + (__int64)iDistance);
}


int WinFile::Read(void *lpBuffer, int iBytes)
{
	// Read bytes from the file.

	unsigned char *dst;
	int remains;
	DWORD offs;

	int ret = 0;

	if (m_Handle == INVALID_HANDLE_VALUE) goto Abort;
	if (lpBuffer == NULL) goto Abort;
	if (iBytes <= 0) goto Abort;

	// Don't read past EOF
	if (m_Pos.QuadPart >= m_Size.QuadPart) goto Abort;

	if ((LONGLONG)iBytes > (m_Size.QuadPart - m_Pos.QuadPart))
	{
		iBytes = (int)(m_Size.QuadPart - m_Pos.QuadPart);
	}

	dst = (unsigned char *)lpBuffer;

	// Start by reading all bytes up to the next page transition

	offs = m_Pos.LowPart & (WF_BUFFERSIZE - 1);
	remains = WF_BUFFERSIZE - offs;

	for (;;) {
		if (remains > iBytes) remains = iBytes;
		CopyMemory(dst, m_Buff + offs, remains);

		m_Pos.QuadPart += remains;
		ret += remains;
		iBytes -= remains;

		// If no bytes remain, we're finished
		if (iBytes <= 0) break;

		// Otherwise, start reading from a new page
		if (m_Flags & WF_DIRTY)
		{
			m_Flags &= (~WF_DIRTY);
			if (!Flush()) goto Error;
		}
		if (!Refresh()) goto Error;

		dst += remains;
		offs = 0;
		remains = WF_BUFFERSIZE;
	}

	// If we finished exactly on a page boundary, refresh
	if ((m_Pos.LowPart & (WF_BUFFERSIZE - 1)) == 0)
	{
		if (m_Flags & WF_DIRTY)
		{
			m_Flags &= (~WF_DIRTY);
			if (!Flush()) goto Error;
		}
		if (!Refresh()) goto Error;
	}

Abort:
	return ret;

Error:
	Kill();
	return ret;
}


bool WinFile::Write(const void *lpBuffer, int iBytes)
{
	// Write bytes to the file.

	const unsigned char *src;
	int remains;
	DWORD offs;

	bool ret = false;

	if (m_Handle == INVALID_HANDLE_VALUE) goto Abort;
	if ((m_Flags & WF_WRITE) == 0) goto Abort;

	if (lpBuffer == NULL) goto Abort;
	if (iBytes < 0) goto Abort;
	if (iBytes == 0) return true;

	src = (const unsigned char *)lpBuffer;

	// Start by writing all bytes up to the next page transition

	offs = m_Pos.LowPart & (WF_BUFFERSIZE - 1);
	remains = WF_BUFFERSIZE - offs;

	// Buffer will always be dirty during the loop
	m_Flags |= WF_DIRTY;

	for (;;) {
		if (remains > iBytes) remains = iBytes;
		CopyMemory(m_Buff + offs, src, remains);

		m_Pos.QuadPart += remains;
		iBytes -= remains;

		if (m_Pos.QuadPart > m_Size.QuadPart)
		{
			m_Size.QuadPart = m_Pos.QuadPart;
		}

		// If no bytes remain, we're finished
		if (iBytes <= 0) break;

		// Otherwise, start writing from a new page
		if (!Flush()) goto Error;
		if (!Refresh()) goto Error;

		src += remains;
		offs = 0;
		remains = WF_BUFFERSIZE;
	}

	// If we finished exactly on a page boundary, refresh.
	// (This is the only case in which the buffer will not
	// be dirty when we leave this function.)

	ret = ((m_Pos.LowPart & (WF_BUFFERSIZE - 1)) != 0);

	if (!ret)
	{
		if (!Flush()) goto Error;
		m_Flags &= (~WF_DIRTY);

		if (!Refresh()) goto Error;
		ret = true;
	}

Abort:
	return ret;

Error:
	Kill();
	return false;
}


int WinFile::ReadByte()
{
	// Read one byte from the file.

	int ret = 0;

	if (m_Handle != INVALID_HANDLE_VALUE)
	{
		if (m_Pos.QuadPart < m_Size.QuadPart)
		{
			DWORD offs = m_Pos.LowPart & (WF_BUFFERSIZE - 1);

			ret = m_Buff[offs];
			++m_Pos.QuadPart;

			if (offs == (WF_BUFFERSIZE - 1))
			{
				if (m_Flags & WF_DIRTY)
				{
					m_Flags &= (~WF_DIRTY);
					if (!Flush()) goto Error;
				}
				if (!Refresh()) goto Error;
			}
		}
	}

	return ret;

Error:
	Kill();
	return ret;
}


bool WinFile::WriteByte(int iByte)
{
	// Write one byte to the file.

	bool ret = false;

	if (m_Handle != INVALID_HANDLE_VALUE)
	{
		if (m_Flags & WF_WRITE)
		{
			DWORD offs = m_Pos.LowPart & (WF_BUFFERSIZE - 1);

			m_Buff[offs] = (unsigned char)iByte;
			m_Flags |= WF_DIRTY;

			++m_Pos.QuadPart;
			if (m_Pos.QuadPart > m_Size.QuadPart)
			{
				m_Size.QuadPart = m_Pos.QuadPart;
			}

			ret = (offs != (WF_BUFFERSIZE - 1));
			if (!ret)
			{
				if (!Flush()) goto Error;
				m_Flags &= (~WF_DIRTY);

				if (!Refresh()) goto Error;
			}
		}
	}

	return ret;

Error:
	Kill();
	return false;
}


void WinFile::AlignToByte()
{
	// Align at the next byte boundary

	m_CurByte = 0;
	m_CurBit  = 0;
}


int WinFile::GetBit()
{
	// Read one bit from the file

	int ret;

	if (m_CurBit == 0)
	{
		m_CurByte = ReadByte();
		m_CurBit  = 128;
	}

	ret = (m_CurByte & m_CurBit)? 1: 0;
	m_CurBit >>= 1;

	return ret;
}


int WinFile::GetBits(int nBits)
{
	// Read multiple bits from the file

	int ret = 0;

	while (nBits > 0)
	{
		ret = ret + ret + GetBit();
		--nBits;
	}

	return ret;
}

